# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Object detection class using YOLOv4 single label model to find license
plates.
"""

import logging
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import cv2
import numpy as np
import tensorflow as tf
from tensorflow.python.saved_model import tag_constants


class Detector:  # pylint: disable=too-many-instance-attributes
    """Object detection class using yolo model to find object bboxes"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        class_names: List[str],
        model_type: str,
        model_file: Dict[str, str],
        max_output_size_per_class: int,
        max_total_size: int,
        input_size: int,
        iou_threshold: float,
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.class_names = class_names
        self.model_type = model_type
        self.model_path = model_dir / model_file[self.model_type]

        self.max_output_size_per_class = max_output_size_per_class
        self.max_total_size = max_total_size
        self.input_size = (input_size, input_size)
        self.iou_threshold = iou_threshold
        self.score_threshold = score_threshold

        self.yolo = self._create_yolo_model()

    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Detects all license plate objects' bounding box from one image

        Args:
            image (np.ndarray): Input image.

        Return:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): A tuple containing the
            following arrays:
            - boxes: An array of bounding box with definition like
                (x1, y1, x2, y2), in a coordinate system with origin point in
                the top-left corner
            - labels: An array of class labels.
            - scores: An array of confidence scores.
        """
        image = self._preprocess(image)

        pred = self.yolo(tf.constant(image))
        pred = next(iter(pred.values()))

        bboxes, scores, classes = self._postprocess(pred[:, :, :4], pred[:, :, 4:])
        labels = np.array([self.class_names[int(i)] for i in classes])

        return bboxes, labels, scores

    def _create_yolo_model(self) -> Callable:
        """Creates yolo model for license plate detection."""
        self.logger.info(
            "Yolo model loaded with following configs:\n\t"
            f"Model type: {self.model_type},\n\t"
            f"Input resolution: {self.input_size},\n\t"
            f"Max detections per class: {self.max_output_size_per_class},\n\t"
            f"Max total detections: {self.max_total_size},\n\t"
            f"IOU threshold: {self.iou_threshold},\n\t"
            f"Score threshold: {self.score_threshold}"
        )

        return self._load_yolo_weights()

    def _load_yolo_weights(self) -> Callable:
        self.model = tf.saved_model.load(
            str(self.model_path), tags=[tag_constants.SERVING]
        )
        return self.model.signatures["serving_default"]

    def _postprocess(
        self,
        pred_boxes: tf.Tensor,
        pred_scores: tf.Tensor,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        # performs nms using model's predictions
        bboxes, scores, classes, valid_dets = tf.image.combined_non_max_suppression(
            tf.reshape(pred_boxes, (tf.shape(pred_boxes)[0], -1, 1, 4)),
            tf.reshape(
                pred_scores, (tf.shape(pred_scores)[0], -1, tf.shape(pred_scores)[-1])
            ),
            self.max_output_size_per_class,
            self.max_total_size,
            self.iou_threshold,
            self.score_threshold,
        )
        num_valid = valid_dets[0]

        classes = classes.numpy()[0]
        classes = classes[:num_valid]

        scores = scores.numpy()[0]
        scores = scores[:num_valid]

        bboxes = bboxes.numpy()[0]
        bboxes = bboxes[:num_valid]

        # swapping x and y axes
        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]

        # scaling of bboxes if v4tiny model is used
        if self.model_type == "v4tiny":
            bboxes = self.scale_bboxes(bboxes, 0.75)

        return bboxes, scores, classes

    def _preprocess(self, image: np.ndarray) -> np.ndarray:
        image = cv2.resize(image, self.input_size)
        image = np.asarray([image]).astype(np.float32) / 255.0

        return image

    @staticmethod
    def scale_bboxes(bboxes: np.ndarray, scale_factor: float) -> np.ndarray:
        """Scales the width and height of bboxes from v4tiny.

        After converting the model from .cfg and .weight file format (Alexey's
        Darknet repo) to tf model, bboxes are bigger. So downscaling is
        required for a better fit.
        """
        outputs = np.empty_like(bboxes)
        center_x = (bboxes[:, 0] + bboxes[:, 2]) / 2
        center_y = (bboxes[:, 1] + bboxes[:, 3]) / 2
        outputs[:, 0] = center_x - (bboxes[:, 2] - bboxes[:, 0]) / 2 * scale_factor
        outputs[:, 1] = center_y - (bboxes[:, 3] - bboxes[:, 1]) / 2 * scale_factor
        outputs[:, 2] = center_x + (bboxes[:, 2] - bboxes[:, 0]) / 2 * scale_factor
        outputs[:, 3] = center_y + (bboxes[:, 3] - bboxes[:, 1]) / 2 * scale_factor
        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Predictor class to handle detection of poses for movenet
"""

import logging
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import cv2
import numpy as np
import tensorflow as tf
from tensorflow.python.saved_model import tag_constants

# fmt: off
SKELETON = [
    [16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13], [6, 7],
    [6, 8], [7, 9], [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4], [3, 5],
    [4, 6], [5, 7],
]
# fmt: on


class Predictor:  # pylint: disable=too-many-instance-attributes
    """Predictor class to handle detection of poses for MoveNet."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        model_type: str,
        model_file: Dict[str, str],
        resolution: Dict[str, Dict[str, int]],
        bbox_score_threshold: float,
        keypoint_score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.model_type = model_type
        self.model_path = model_dir / model_file[self.model_type]
        self.resolution = self.get_resolution_as_tuple(resolution[self.model_type])

        self.bbox_score_threshold = bbox_score_threshold
        self.keypoint_score_threshold = keypoint_score_threshold

        self.movenet = self._create_movenet_model()

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # pylint: disable=too-many-locals
        """MoveNet prediction function

        Args:
            frame (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): Nx4 array of bboxes, N is number of detections
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates
            keypoints_scores (np.ndarray): Nx17 array of keypoint scores
            keypoints_conns (np.ndarray): NxD'x2 keypoint connections, where
                D' is the varying pairs of valid keypoint connections per detection
        """
        image_data = cv2.resize(frame, (self.resolution))
        image_data = np.asarray([image_data]).astype(np.int32)
        outputs = self.movenet(tf.constant(image_data))
        predictions = outputs["output_0"]

        if "multi" in self.model_type:
            (
                bboxes,
                keypoints,
                keypoints_scores,
                keypoints_conns,
            ) = self._get_results_multi(predictions)
        else:
            (
                bboxes,
                keypoints,
                keypoints_scores,
                keypoints_conns,
            ) = self._get_results_single(predictions)

        return bboxes, keypoints, keypoints_scores, keypoints_conns

    def _create_movenet_model(self) -> Callable:
        """Creates the MoveNet model."""
        # movenet singlepose do not output bbox, so bbox score threshold not
        # applicable
        bbox_score_threshold = (
            self.bbox_score_threshold
            if "multi" in self.model_type
            else "NA for singlepose models"
        )
        self.logger.info(
            f"MoveNet model loaded with following configs:\n\t"
            f"Model type: {self.model_type},\n\t"
            f"Input resolution: {self.resolution},\n\t"
            f"bbox_score_threshold: {bbox_score_threshold},\n\t"
            f"keypoint_score_threshold: {self.keypoint_score_threshold}"
        )

        return self._load_movenet_weights()

    def _get_results_multi(
        self, predictions: tf.Tensor
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns formatted outputs for multipose model.

        Predictions come in a [1x6x56] tensor for up to 6 people detections.
        First 17 x 3 = 51 elements are keypoints locations in the form of
        [y_0, x_0, s_0, y_1, x_1, s_1, …, y_16, x_16, s_16],
        where y_i, x_i, s_i are the yx-coordinates and confidence score.
        Remaining 5 elements [ymin, xmin, ymax, xmax, score] represent
        bbox coordinates, and confidence score

        Args:
            predictions (tf.Tensor): Model output in a [1x6x56] tensor.

        Returns:
            bboxes (np.ndarray): Nx4 array of bboxes, N is number of detections.
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates.
            keypoints_scores (np.ndarray): Nx17 array of keypoint scores.
            keypoints_conns (np.ndarray): NxD'x2 keypoint connections, where
                D' is the varying pair of valid keypoint connections per
                detection.
        """
        predictions = tf.squeeze(predictions, axis=0).numpy()
        keypoints_x = predictions[:, range(1, 51, 3)]
        keypoints_y = predictions[:, range(0, 51, 3)]
        keypoints_scores = predictions[:, range(2, 51, 3)]
        bboxes = predictions[:, 51:55]
        bbox_score = predictions[:, 55]
        keypoints = (
            tf.concat(
                [
                    tf.expand_dims(keypoints_x, axis=2),
                    tf.expand_dims(keypoints_y, axis=2),
                ],
                2,
            )
        ).numpy()
        # swap bbox coordinates from y1,x1,y2,x2 to x1,y1,x2,y2
        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]
        bbox_masks = self._get_masks_from_bbox_scores(
            bbox_score, self.bbox_score_threshold
        )
        if len(bbox_masks) == 0:
            return (np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0))

        bboxes = bboxes[bbox_masks, :]
        bbox_score = bbox_score[bbox_masks]
        keypoints = keypoints[bbox_masks, :, :]
        keypoints_scores = keypoints_scores[bbox_masks, :]
        valid_keypoints, keypoints_masks = self._get_keypoints_coords(
            keypoints, keypoints_scores, self.keypoint_score_threshold
        )
        keypoints_conns = self._get_connections_of_poses(keypoints, keypoints_masks)

        return bboxes, valid_keypoints, keypoints_scores, keypoints_conns

    def _get_results_single(
        self, predictions: tf.Tensor
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Returns formatted outputs for singlepose model.

        Predictions come in a [1x1x17x3] tensor.
        First 2 channel in the last dimension represent the (y,x) coordinates,
        the 3rd channel represents confidence scores for the keypoints.

        Args:
            predictions (tf.Tensor): Model output in a [1x1x17x3] tensor.

        Returns:
            bboxes (np.ndarray): 1x4 array of bboxes.
            keypoints (np.ndarray): 1x17x2 array of keypoint coordinates.
            keypoints_scores (np.ndarray): 1x17 array of keypoint scores.
            keypoints_conns (np.ndarray): 1xD'x2 keypoint connections, where
                D' is the varying pair of valid keypoint connections per
                detection.
        """
        predictions = tf.squeeze(predictions, axis=0)
        predictions = tf.squeeze(predictions, axis=0).numpy()
        keypoints_x = predictions[:, 1]
        keypoints_y = predictions[:, 0]
        keypoints_scores = predictions[:, 2].reshape((1, -1))

        # to concatenate the xy coordinate together for each keypoint
        keypoints = (
            tf.concat(
                [
                    tf.expand_dims(keypoints_x, axis=1),
                    tf.expand_dims(keypoints_y, axis=1),
                ],
                1,
            )
        ).numpy()
        keypoints = keypoints.reshape((1, -1, 2))
        valid_keypoints, keypoints_masks = self._get_keypoints_coords(
            keypoints, keypoints_scores, self.keypoint_score_threshold
        )
        # if valid_keypoints is None:
        #     return (np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0))
        # if all the keypoints_score are below threshold, valid_keypoints
        # will be an array with all [-1,-1]
        if np.all(valid_keypoints == np.asarray([[-1, -1]])):
            return (np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0))

        keypoints_conns = self._get_connections_of_poses(keypoints, keypoints_masks)
        bbox = self._get_bbox_from_keypoints(valid_keypoints)

        return bbox, valid_keypoints, keypoints_scores, keypoints_conns

    def _load_movenet_weights(self) -> Callable:
        self.model = tf.saved_model.load(
            str(self.model_path), tags=[tag_constants.SERVING]
        )
        return self.model.signatures["serving_default"]

    @staticmethod
    def get_resolution_as_tuple(resolution: Dict[str, int]) -> Tuple[int, int]:
        """Convert resolution from dict to tuple format

        Args:
            resolution (Dict[str, int]): height and width in dict format

        Returns:
            resolution (Tuple(int)): height and width in tuple format
        """
        return int(resolution["height"]), int(resolution["width"])

    @staticmethod
    def _get_bbox_from_keypoints(valid_keypoints: np.ndarray) -> np.ndarray:
        """Obtains coordinates from the keypoints for single pose model where
        bounding box coordinates are not provided as model outputs. The bounding
        box coordinates will be [xmin,ymin,xmax,ymax], derived from the keypoints
        coordinates.

        Args:
            valid_keypoints (np.ndarray): Nx17x2 array of keypoints where N is
                the number of detections.

        Returns:
            bboxes (np.ndarray): Nx4 array of bbox coordinates in the form of
                [x1,y1,x2,y2] where the values are in relative coordinates
                between 0 and 1.
        """
        # using absolute because it is coded that invalid keypoints are set as -1
        xmin = np.abs(valid_keypoints[:, :, 0]).min()
        ymin = np.abs(valid_keypoints[:, :, 1]).min()
        xmax = valid_keypoints[:, :, 0].max()
        ymax = valid_keypoints[:, :, 1].max()
        bboxes = np.asarray([[xmin, ymin, xmax, ymax]])

        return bboxes

    @staticmethod
    def _get_connections_of_poses(
        keypoints: np.ndarray, masks: np.ndarray
    ) -> np.ndarray:
        """Gets connections between adjacent keypoint pairs if both keypoints are
        detected. Output will be in the shape of NxD'x2 keypoint connections,
        where D' is the varying pairs of valid keypoint connections per
        detection.

        Args:
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates, where N
                is the number of detections.
            masks (np.ndarray): Nx17 boolean for valid keypoints.

        Returns:
            keypoint_conns (np.ndarray): NxD'x2 keypoint connections, where D' is
                the varying pair of valid keypoint connections per detection
        """
        all_connections = []
        for i in range(keypoints.shape[0]):
            connections = []
            for start_joint, end_joint in SKELETON:
                if masks[i, start_joint - 1] and masks[i, end_joint - 1]:
                    connections.append(
                        np.vstack(
                            (keypoints[i, start_joint - 1], keypoints[i, end_joint - 1])
                        )
                    )
            all_connections.append(connections)
        keypoint_conns = np.asarray(all_connections)

        return keypoint_conns

    @staticmethod
    def _get_keypoints_coords(
        keypoints: np.ndarray,
        keypoints_score: np.ndarray,
        keypoints_score_threshold: float,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Keeps only valid keypoints' relative coordinates.

        Args:
            keypoints (np.ndarray): Nx17x2 array of keypoints' relative
                coordinates.
            keypoints_score (np.ndarray): Nx17 keypoints confidence score for
                valid keypoints coordinates.
            keypoints_score_threshold (float): threshold for keypoints score.

        Returns:
            valid_keypoints (np.ndarray): Nx17x2 array of keypoints where
                undetected keypoints are assigned a (-1,-1) value.
            keypoint_masks (np.ndarray) : Nx17 boolean for valid keypoints
        """
        valid_keypoints = keypoints.copy()
        valid_keypoints = np.clip(valid_keypoints, 0, 1)
        keypoint_masks = keypoints_score > keypoints_score_threshold
        valid_keypoints[~keypoint_masks] = [-1, -1]

        return valid_keypoints, keypoint_masks

    @staticmethod
    def _get_masks_from_bbox_scores(
        bbox_scores: np.ndarray, score_threshold: float
    ) -> List[int]:
        """Obtains masks for bbox with confidence scores above the detection
        threshold. The mask come in the format of a list with idx of approved
        detection, e.g., [0,1,2], generally movenet will have the valid
        detections in the first few indexes

        Args:
            bbox_scores (np.ndarray): Array of bbox confidence scores.
            score_threshold (float): Threshold value for bbox confidence scores.

        Returns:
            masks (list): List of index of bboxes that have respective bbox
                confidence scores above the score threshold.
        """
        bbox_scores_list = bbox_scores.tolist()
        masks = [
            idx for idx, score in enumerate(bbox_scores_list) if score > score_threshold
        ]
        return masks



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Assigns objects in close proximity to groups.
"""

from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.dabble.utils.quick_find import QuickFind
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Groups objects that are near each other.

    It does so by comparing the 3D locations of all objects, and assigning
    objects near each other to the same group. The group associated with each
    object is accessed by the ``groups`` key of :term:`obj_attrs`.

    Inputs:
        |obj_3D_locs_data|

    Outputs:
        |obj_attrs_data|
        :mod:`dabble.group_nearby_objs` produces the ``groups`` attribute.

    Configs:
        obj_dist_threshold (:obj:`float`): **default = 1.5**. |br|
            Threshold of distance, in metres, between two objects. Objects with
            distance less than ``obj_dist_threshold`` would be assigned to the
            same group.

    .. versionchanged:: 1.2.0
        :mod:`draw.group_nearby_objs` used to return ``obj_tags``
        (:obj:`List[str]`) as an output data type, which has been deprecated
        and now subsumed under :term:`obj_attrs`. The same attribute is
        accessed by the ``groups`` key of :term:`obj_attrs`.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Checks the distance between 3d locations of a pair of objects.

        If distance is less than threshold, assign the objects to the same
        group. Repeat for all object pairs.
        """
        nearby_obj_pairs = self._find_nearby_obj_pairs(
            inputs["obj_3D_locs"], self.obj_dist_threshold
        )

        quickfind = QuickFind(len(inputs["obj_3D_locs"]))
        for (idx_1, idx_2) in nearby_obj_pairs:
            if not quickfind.connected(idx_1, idx_2):
                quickfind.union(idx_1, idx_2)

        return {"obj_attrs": {"groups": quickfind.get_group_alloc()}}

    @staticmethod
    def _find_nearby_obj_pairs(
        obj_locs: List[np.ndarray], obj_dist_threshold: float
    ) -> List[Tuple[int, int]]:
        """If the distance between 2 objects are less than the threshold,
        append their indexes to nearby_obj_pairs as a tuple.
        """
        nearby_obj_pairs = []
        for idx_1, loc_1 in enumerate(obj_locs):
            for idx_2, loc_2 in enumerate(obj_locs):
                if idx_1 == idx_2:
                    continue

                dist_bet = np.linalg.norm(loc_1 - loc_2)

                if dist_bet <= obj_dist_threshold:
                    if (idx_2, idx_1) not in nearby_obj_pairs:
                        nearby_obj_pairs.append((idx_1, idx_2))

        return nearby_obj_pairs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""YOLOX models with model types: yolox-tiny, yolox-s, yolox-m, and yolox-l."""

import logging
from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.detector import Detector


class YOLOXModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """Validates configuration, loads YOLOX model, and performs inference.

    Configuration options are validated to ensure they have valid types and
    values. Model weights files are downloaded if not found in the location
    indicated by the `weights_dir` configuration option.

    Attributes:
        class_names (List[str]): Human-friendly class names of the object
            categories.
        detect_ids (List[int]): List of selected object category IDs. IDs not
            found in this list will be filtered away from the results. An empty
            list indicates that all object categories should be detected.
        detector (Detector): YOLOX detector object to infer bboxes from a
            provided image frame.
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_bounds(["iou_threshold", "score_threshold"], "[0, 1]")

        model_dir = self.download_weights()
        with open(model_dir / self.weights["classes_file"]) as infile:
            class_names = [line.strip() for line in infile.readlines()]

        self.detect_ids = self.config["detect_ids"]
        self.detector = Detector(
            model_dir,
            class_names,
            self.detect_ids,
            self.config["model_type"],
            self.config["num_classes"],
            self.config["model_size"],
            self.weights["model_file"],
            self.config["agnostic_nms"],
            self.config["fuse"],
            self.config["half"],
            self.config["input_size"],
            self.config["iou_threshold"],
            self.config["score_threshold"],
        )

    @property
    def detect_ids(self) -> List[int]:
        """The list of selected object category IDs."""
        return self._detect_ids

    @detect_ids.setter
    def detect_ids(self, ids: List[int]) -> None:
        if not isinstance(ids, list):
            raise TypeError("detect_ids has to be a list")
        self._detect_ids = ids

    def predict(self, image: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts bboxes from image.

        Args:
            image (np.ndarray): Input image frame.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Returned tuple
            contains:
            - An array of detection bboxes
            - An array of human-friendly detection class names
            - An array of detection scores

        Raises:
            TypeError: The provided `image` is not a numpy array.
        """
        if not isinstance(image, np.ndarray):
            raise TypeError("image must be a np.ndarray")
        return self.detector.predict_object_bbox_from_image(image)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mixin classes for PeekingDuck nodes and models."""

import hashlib
import operator
import os
import re
import sys
import zipfile
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Union

import requests
from tqdm import tqdm

BASE_URL = "https://storage.googleapis.com/peekingduck/models"
PEEKINGDUCK_WEIGHTS_SUBDIR = "peekingduck_weights"


class ThresholdCheckerMixin:
    """Mixin class providing utility methods for checking validity of config
    values, typically thresholds.
    """

    interval_pattern = re.compile(
        r"^[\[\(]\s*[-+]?(inf|\d*\.?\d+)\s*,\s*[-+]?(inf|\d*\.?\d+)\s*[\]\)]$"
    )

    def check_bounds(self, key: Union[str, List[str]], interval: str) -> None:
        """Checks if the configuration value(s) specified by `key` satisfies
        the specified bounds.

        Args:
            key (Union[str, List[str]]): The specified key or list of keys.
            interval (str): An mathematical interval representing the range of
                valid values. The syntax of the `interval` string is:

                <value> = <number> | "-inf" | "+inf"
                <left_bracket> = "(" | "["
                <right_bracket> = ")" | "]"
                <interval> = <left_bracket> <value> "," <value> <right_bracket>

                See Technotes for more details.

        Raises:
            TypeError: `key` type is not in (List[str], str).
            ValueError: If `interval` does not match the specified format.
            ValueError: If the lower bound is larger than the upper bound.
            ValueError: If the configuration value fails the bounds comparison.

        Technotes:
            The table below shows the comparison done for various interval
            expressions.

            +---------------------+-------------------------------------+
            | interval            | comparison                          |
            +=====================+=====================================+
            | [lower, +inf]       |                                     |
            +---------------------+                                     |
            | [lower, +inf)       | config[key] >= lower                |
            +---------------------+-------------------------------------+
            | (lower, +inf]       |                                     |
            +---------------------+                                     |
            | (lower, +inf)       | config[key] > lower                 |
            +---------------------+-------------------------------------+
            | [-inf, upper]       |                                     |
            +---------------------+                                     |
            | (-inf, upper]       | config[key] <= upper                |
            +---------------------+-------------------------------------+
            | [-inf, upper)       |                                     |
            +---------------------+                                     |
            | (-inf, upper)       | config[key] < upper                 |
            +---------------------+-------------------------------------+
            | [lower, upper]      | lower <= config[key] <= upper       |
            +---------------------+-------------------------------------+
            | (lower, upper]      | lower < config[key] <= upper        |
            +---------------------+-------------------------------------+
            | [lower, upper)      | lower <= config[key] < upper        |
            +---------------------+-------------------------------------+
            | (lower, upper)      | lower < config[key] < upper         |
            +---------------------+-------------------------------------+
        """
        if self.interval_pattern.match(interval) is None:
            raise ValueError("Badly formatted interval")

        left_bracket = interval[0]
        right_bracket = interval[-1]
        lower, upper = [float(value.strip()) for value in interval[1:-1].split(",")]

        if lower > upper:
            raise ValueError("Lower bound cannot be larger than upper bound")

        self._check_within_bounds(key, lower, upper, left_bracket, right_bracket)

    def check_valid_choice(
        self, key: str, choices: Set[Union[int, float, str]]
    ) -> None:
        """Checks that configuration value specified by `key` can be found
        in `choices`.

        Args:
            key (str): The specified key.
            choices (Set[Union[int, float, str]]): The valid choices.

        Raises:
            TypeError: `key` type is not a str.
            ValueError: If the configuration value is not found in `choices`.
        """
        if not isinstance(key, str):
            raise TypeError("`key` must be str")
        if self.config[key] not in choices:
            raise ValueError(f"{key} must be one of {choices}")

    def _check_within_bounds(  # pylint: disable=too-many-arguments
        self,
        key: Union[str, List[str]],
        lower: float,
        upper: float,
        left_bracket: str,
        right_bracket: str,
    ) -> None:
        """Checks that configuration values specified by `key` is within the
        specified bounds between `lower` and `upper`.

        Args:
            key (Union[str, List[str]]): The specified key or list of keys.
            lower (float): The lower bound.
            upper (float): The upper bound.
            left_bracket (str): Either a "(" for an open lower bound or a "["
                for a closed lower bound.
            right_bracket (str): Either a ")" for an open upper bound or a "]"
                for a closed upper bound.

        Raises:
            TypeError: `key` type is not in (List[str], str).
            ValueError: If the configuration value is not between `lower` and
                `upper`.
        """
        method_lower = operator.ge if left_bracket == "[" else operator.gt
        method_upper = operator.le if right_bracket == "]" else operator.lt
        reason = f"between {left_bracket}{lower}, {upper}{right_bracket}"
        self._compare(key, lower, method_lower, reason)
        self._compare(key, upper, method_upper, reason)

    def _compare(
        self,
        key: Union[str, List[str]],
        value: Union[float, int],
        method: Callable,
        reason: str,
    ) -> None:
        """Compares the configuration values specified by `key` with
        `value` using the specified comparison `method`, raises error with
        `reason` if comparison fails.

        Args:
            key (Union[str, List[str]]): The specified key or list of keys.
            value (Union[float, int]): The specified value.
            method (Callable): The method to be used to compare the
                configuration value specified by `key` and `value`.
            reason (str): The failure reason.

        Raises:
            TypeError: `key` type is not in (List[str], str).
            ValueError: If the comparison between `config[key]` and `value`
                fails.
        """
        if isinstance(key, str):
            if isinstance(self.config[key], list):
                if not all(method(val, value) for val in self.config[key]):
                    raise ValueError(f"All elements of {key} must be {reason}")
            elif not method(self.config[key], value):
                raise ValueError(f"{key} must be {reason}")
        elif isinstance(key, list):
            for k in key:
                self._compare(k, value, method, reason)
        else:
            raise TypeError("`key` must be either str or list")


class WeightsDownloaderMixin:
    """Mixin class providing utility methods for downloading model weights."""

    @property
    def blob_filename(self) -> str:
        """Name of the selected weights on GCP."""
        return self.weights["blob_file"][self.config["model_type"]]

    @property
    def classes_filename(self) -> Optional[str]:
        """Name of the file containing classes IDs/labels for the selected
        model.
        """
        return self.weights.get("classes_file")

    @property
    def weights(self) -> Dict[str, Any]:
        """Dictionary of `blob_file`, `config_file`, and `model_file` names
        based on the selected `model_format`.
        """
        return self.config["weights"][self.config["model_format"]]

    @property
    def model_filename(self) -> str:
        """Name of the selected weights on local machine."""
        return self.weights["model_file"][self.config["model_type"]]

    @property
    def model_subdir(self) -> str:
        """Model weights sub-directory name based on the selected model
        format.
        """
        return self.weights["model_subdir"]

    def download_weights(self) -> Path:
        """Downloads weights for specified ``blob_file``.

        Returns:
            (Path): Path to the directory where the model's weights are stored.
        """
        model_dir = self._find_paths()
        if self._has_weights(model_dir):
            return model_dir

        self.logger.info("Proceeding to download...")

        model_dir.mkdir(parents=True, exist_ok=True)
        self._download_to(self.blob_filename, model_dir)
        self._extract_file(model_dir)
        if self.classes_filename is not None:
            self._download_to(self.classes_filename, model_dir)

        self.logger.info(f"Weights downloaded to {model_dir}.")

        return model_dir

    def _download_to(self, filename: str, destination_dir: Path) -> None:
        """Downloads publicly shared files from Google Cloud Platform.

        Saves download content in chunks. Chunk size set to large integer as
        weights are usually pretty large.

        Args:
            destination_dir (Path): Destination directory of downloaded file.
        """
        with open(destination_dir / filename, "wb") as outfile, requests.get(
            f"{BASE_URL}/{self.model_subdir}/{self.config['model_format']}/{filename}",
            stream=True,
        ) as response:
            for chunk in tqdm(response.iter_content(chunk_size=32768)):
                if chunk:  # filter out keep-alive new chunks
                    outfile.write(chunk)

    def _extract_file(self, destination_dir: Path) -> None:
        """Extracts the zip file to ``destination_dir``.

        Args:
            destination_dir (Path): Destination directory for extraction.
        """
        zip_path = destination_dir / self.blob_filename
        with zipfile.ZipFile(zip_path, "r") as infile:
            file_list = infile.namelist()
            for file in tqdm(file=sys.stdout, iterable=file_list, total=len(file_list)):
                infile.extract(member=file, path=destination_dir)

        os.remove(zip_path)

    def _find_paths(self) -> Path:
        """Constructs the `peekingduck_weights` directory path and the model
        sub-directory path.

        Returns:
            (Path): /path/to/peekingduck_weights/<model_name> where
              weights for a model are stored.

        Raises:
            FileNotFoundError: When the user-specified `weights_parent_dir`
                does not exist.
            ValueError: When the user-specified `weights_parent_dir` is not an
                absolute path.
        """
        if self.config["weights_parent_dir"] is None:
            weights_parent_dir = self.config["root"].parent
        else:
            weights_parent_dir = Path(self.config["weights_parent_dir"])

            if not weights_parent_dir.exists():
                raise FileNotFoundError(
                    f"weights_parent_dir does not exist: {weights_parent_dir}"
                )
            if not weights_parent_dir.is_absolute():
                raise ValueError(
                    f"weights_parent_dir must be an absolute path: {weights_parent_dir}"
                )

        return (
            weights_parent_dir
            / PEEKINGDUCK_WEIGHTS_SUBDIR
            / self.model_subdir
            / self.config["model_format"]
        )

    def _get_weights_checksum(self) -> str:
        with requests.get(f"{BASE_URL}/weights_checksums.json") as response:
            checksums = response.json()
        return checksums[self.model_subdir][self.config["model_format"]][
            str(self.config["model_type"])
        ]

    def _has_weights(self, model_dir: Path) -> bool:
        """Checks if the specified weights file is present in the model
        sub-directory of the PeekingDuck weights directory.

        Args:
            model_dir (Path): /path/to/peekingduck_weights/<model_name> where
                weights for a model are stored.

        Returns:
            (bool): ``True`` if specified weights file in ``model_dir``
            exists and up-to-date/not corrupted, else ``False``.
        """
        weights_path = model_dir / self.model_filename
        if not weights_path.exists():
            self.logger.warning("No weights detected.")
            return False
        if self.sha256sum(weights_path).hexdigest() != self._get_weights_checksum():
            self.logger.warning("Weights file is corrupted/out-of-date.")
            return False
        return True

    @staticmethod
    def sha256sum(path: Path, hash_func: "hashlib._Hash" = None) -> "hashlib._Hash":
        """Hashes the specified file/directory using SHA256. Reads the file in
        chunks to be more memory efficient.

        When a directory path is passed as the argument, sort the folder
        content and hash the content recursively.

        Args:
            path (Path): Path to the file to be hashed.
            hash_func (Optional[hashlib._Hash]): A hash function which uses the
                SHA-256 algorithm.

        Returns:
            (hashlib._Hash): The updated hash function.
        """
        if hash_func is None:
            hash_func = hashlib.sha256()

        if path.is_dir():
            for subpath in sorted(path.iterdir()):
                if subpath.name not in {".DS_Store", "__MACOSX"}:
                    hash_func = WeightsDownloaderMixin.sha256sum(subpath, hash_func)
        else:
            buffer_size = hash_func.block_size * 1024
            with open(path, "rb") as infile:
                for chunk in iter(lambda: infile.read(buffer_size), b""):
                    hash_func.update(chunk)
        return hash_func



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Detector module to predict object bbox from an image using YOLOX."""

import logging
from pathlib import Path
from typing import Dict, List, Tuple

import cv2
import numpy as np
import torch
import torchvision

from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.model import YOLOX
from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.utils import fuse_model
from peekingduck.pipeline.utils.bbox.transforms import xywh2xyxy, xyxy2xyxyn

NUM_CHANNELS = 3


class Detector:  # pylint: disable=too-many-instance-attributes
    """Object detection class using YOLOX to predict object bboxes.

    Attributes:
        logger (logging.Logger): Events logger.
        config (Dict[str, Any]): YOLOX node configuration.
        model_dir (pathlib.Path): Path to directory of model weights files.
        device (torch.device): Represents the device on which the torch.Tensor
            will be allocated.
        half (bool): Flag to determine if half-precision should be used.
        yolox (YOLOX): The YOLOX model for performing inference.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        class_names: List[str],
        detect_ids: List[int],
        model_type: str,
        num_classes: int,
        model_size: Dict[str, Dict[str, float]],
        model_file: Dict[str, str],
        agnostic_nms: bool,
        fuse: bool,
        half: bool,
        input_size: int,
        iou_threshold: float,
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.class_names = class_names
        self.model_type = model_type
        self.num_classes = num_classes
        self.model_size = model_size[self.model_type]
        self.model_path = model_dir / model_file[self.model_type]
        self.agnostic_nms = agnostic_nms
        self.fuse = fuse
        # Half-precision only supported on CUDA
        self.half = half and self.device.type == "cuda"
        self.input_size = (input_size, input_size)
        self.iou_threshold = iou_threshold
        self.score_threshold = score_threshold

        self.update_detect_ids(detect_ids)

        self.yolox = self._create_yolox_model()

    @torch.no_grad()
    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Detects bounding boxes of selected object categories from an image.

        The input image is first scaled according to the `input_size`
        configuration option. Detection results will be filtered according to
        `iou_threshold`, `score_threshold`, and `detect_ids` configuration
        options. Bounding boxes coordinates are then normalized w.r.t. the
        input `image` size.

        Args:
            image (np.ndarray): Input image.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Returned tuple
            contains:
            - An array of detection bboxes
            - An array of human-friendly detection class names
            - An array of detection scores
        """
        # Store the original image size to normalize bbox later
        image_size = image.shape[:2]
        image, scale = self._preprocess(image)
        image = torch.from_numpy(image).unsqueeze(0).to(self.device)
        image = image.half() if self.half else image.float()

        prediction = self.yolox(image)[0]
        bboxes, classes, scores = self._postprocess(
            prediction, scale, image_size, self.class_names
        )

        return bboxes, classes, scores

    def update_detect_ids(self, ids: List[int]) -> None:
        """Updates list of selected object category IDs. When the list is
        empty, all available object category IDs are detected.

        Args:
            ids: List of selected object category IDs
        """
        self.detect_ids = torch.Tensor(ids).to(self.device)  # type: ignore
        if self.half:
            self.detect_ids = self.detect_ids.half()

    def _create_yolox_model(self) -> YOLOX:
        """Creates a YOLOX model and loads its weights.

        Creates `detect_ids` as a `torch.Tensor`. Sets up `input_size` to a
        square shape. Logs model configurations.

        Returns:
            (YOLOX): YOLOX model.
        """
        self.logger.info(
            "YOLOX model loaded with the following configs:\n\t"
            f"Model type: {self.model_type}\n\t"
            f"Input resolution: {self.input_size}\n\t"
            f"IDs being detected: {self.detect_ids.int().tolist()}\n\t"
            f"IOU threshold: {self.iou_threshold}\n\t"
            f"Score threshold: {self.score_threshold}\n\t"
            f"Class agnostic NMS: {self.agnostic_nms}\n\t"
            f"Half-precision floating-point: {self.half}\n\t"
            f"Fuse convolution and batch normalization layers: {self.fuse}"
        )
        return self._load_yolox_weights()

    def _get_model(self, model_size: Dict[str, float]) -> YOLOX:
        """Constructs YOLOX model based on parsed configuration.

        Args:
            model_size (Dict[str, float]): Depth and width of the model.

        Returns:
            (YOLOX): YOLOX model.
        """
        return YOLOX(
            self.num_classes,
            model_size["depth"],
            model_size["width"],
        )

    def _load_yolox_weights(self) -> YOLOX:
        """Loads YOLOX model weights.

        Args:
            model_path (Path): Path to model weights file.
            model_settings (Dict[str, float]): Depth and width of the model.

        Returns:
            (YOLOX): YOLOX model.

        Raises:
            ValueError: `model_path` does not exist.
        """
        if self.model_path.is_file():
            ckpt = torch.load(str(self.model_path), map_location="cpu")
            model = self._get_model(self.model_size).to(self.device)
            if self.half:
                model.half()
            model.eval()
            model.load_state_dict(ckpt["model"])

            if self.fuse:
                model = fuse_model(model)
            return model
        raise ValueError(
            f"Model file does not exist. Please check that {self.model_path} exists."
        )

    def _postprocess(
        self,
        prediction: torch.Tensor,
        scale: float,
        image_shape: Tuple[int, int],
        class_names: List[str],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Postprocesses detection result to be compatible with other nodes.

        Args:
            prediction (torch.Tensor): Raw detections from the YOLOX model.
            scale (float): Scale factor used during preprocessing.
            image_shape (Tuple[int, int]): Image size of the original input
                image.
            class_names (List[str]): List of human-friendly object category
                names.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Returned tuple
            contains:
            - An array of detection bboxes
            - An array of human-friendly detection class names
            - An array of detection scores
        """
        prediction[:, :4] = xywh2xyxy(prediction[:, :4])
        # Get score and class with highest confidence
        class_score, class_pred = torch.max(
            prediction[:, 5 : 5 + self.num_classes], 1, keepdim=True
        )
        # Filter by score_threshold
        conf_mask = (
            prediction[:, 4] * class_score.squeeze() >= self.score_threshold
        ).squeeze()
        # Detections ordered as (x1, y1, x2, y2, obj_conf, class_conf, class_pred)
        detections = torch.cat((prediction[:, :5], class_score, class_pred.float()), 1)
        detections = detections[conf_mask]
        # Early return if all are below score_threshold
        if not detections.size(0):
            return np.empty((0, 4)), np.empty(0), np.empty(0)

        # Class agnostic NMS
        if self.agnostic_nms:
            nms_out_index = torchvision.ops.nms(
                detections[:, :4],
                detections[:, 4] * detections[:, 5],
                self.iou_threshold,
            )
        else:
            nms_out_index = torchvision.ops.batched_nms(
                detections[:, :4],
                detections[:, 4] * detections[:, 5],
                detections[:, 6],
                self.iou_threshold,
            )
        output = detections[nms_out_index]

        # Filter by detect ids
        if self.detect_ids.size(0):
            output = output[torch.isin(output[:, 6], self.detect_ids)]
        output_np = output.cpu().detach().numpy()
        bboxes = xyxy2xyxyn(output_np[:, :4] / scale, *image_shape)
        scores = output_np[:, 4] * output_np[:, 5]
        classes = np.array([class_names[int(i)] for i in output_np[:, 6]])

        return bboxes, classes, scores

    def _preprocess(self, image: np.ndarray) -> Tuple[np.ndarray, float]:
        """Preprocesses the input image for inference.

        The input image is resized to fit the `input_size` specified in the
        configurations. If the image cannot be resized exactly, a gray canvas,
        rgb(114, 114, 114), is created and the resized image is pasted on the
        top left corner of the canvas. rgb(114, 114, 114) is chosen based on
        the value used by the original repository during training.

        Args:
            image (np.ndarray): Input image.

        Returns:
            (Tuple[np.ndarray, float]): A tuple containing the preprocessed
                image and the scale factor used to resize the image. The shape
                of the preprocessed image is (C, H, W).
        """
        # Initialize canvas for padded image as gray
        padded_img = np.full(
            (self.input_size[0], self.input_size[1], NUM_CHANNELS), 114, dtype=np.uint8
        )
        scale = min(
            self.input_size[0] / image.shape[0], self.input_size[1] / image.shape[1]
        )
        scaled_height = int(image.shape[0] * scale)
        scaled_width = int(image.shape[1] * scale)
        resized_img = cv2.resize(
            image,
            (scaled_width, scaled_height),
            interpolation=cv2.INTER_LINEAR,
        ).astype(np.uint8)
        padded_img[:scaled_height, :scaled_width] = resized_img

        # Rearrange from (H, W, C) to (C, H, W)
        padded_img = padded_img.transpose((2, 0, 1))
        padded_img = np.ascontiguousarray(padded_img, dtype=np.float32)
        return padded_img, scale



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Face detection class using MTCNN model to find face bboxes."""

import logging
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import numpy as np
import tensorflow as tf
from peekingduck.pipeline.nodes.model.mtcnnv1.mtcnn_files.graph_functions import (
    load_graph,
)
from peekingduck.pipeline.utils.bbox.transforms import xyxy2xyxyn


class Detector:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Face detection class using MTCNN model to find bboxes and landmarks."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        model_type: str,
        model_file: Dict[str, str],
        model_nodes: Dict[str, List[str]],
        min_size: int,
        scale_factor: float,
        network_thresholds: List[float],
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.model_path = model_dir / model_file[model_type]
        self.model_nodes = model_nodes

        self.min_size = min_size
        self.scale_factor = scale_factor
        self.network_thresholds = network_thresholds
        self.score_threshold = score_threshold

        self.mtcnn = self._create_mtcnn_model()

    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, scores, and landmarks.

        Args:
            image (np.ndarray): Image in numpy array

        Returns:
            bboxes (np.ndarray): Detected bboxes.
            scores (np.ndarray): Confidence scores.
            landmarks (np.ndarray): Facial landmarks.
        """
        image = self._preprocess(image)
        bboxes, scores, landmarks = self.mtcnn(
            image, self.min_size, self.scale_factor, self.network_thresholds
        )
        bboxes, scores, landmarks = self._post_process(image, bboxes, scores, landmarks)

        return bboxes, scores, landmarks

    def _create_mtcnn_model(self) -> Callable:
        """Creates MTCNN model for face detection."""
        self.logger.info(
            "MTCNN model loaded with following configs:\n\t"
            f"Min size: {self.min_size},\n\t"
            f"Scale Factor: {self.scale_factor},\n\t"
            f"Network Thresholds: {self.network_thresholds},\n\t"
            f"Score Threshold: {self.score_threshold}"
        )

        return self._load_mtcnn_weights()

    def _load_mtcnn_weights(self) -> Callable:
        if not self.model_path.is_file():
            raise ValueError(
                f"Graph file does not exist. Please check that {self.model_path} exists"
            )

        return load_graph(
            str(self.model_path),
            self.model_nodes["inputs"],
            self.model_nodes["outputs"],
        )

    def _post_process(
        self,
        image: np.ndarray,
        bboxes: tf.Tensor,
        scores: tf.Tensor,
        landmarks: tf.Tensor,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Processes MTCNN model outputs. Filter detections by confidence score
        and swaps the x and y coordinates of the bboxes.

        Args:
            image (np.ndarray): Image in numpy array.
            bboxes (tf.Tensor): Tensor array of detected bboxes.
            scores (tf.Tensor): Tensor array of confidence scores.
            landmarks (tf.Tensor): Tensor array of facial landmarks.

        Returns:
            bboxes (np.ndarray): Processed detected bboxes.
            scores (np.ndarray): Processed confidence scores.
            landmarks (np.ndarray): Processed facial landmarks.
        """
        bboxes, scores, landmarks = bboxes.numpy(), scores.numpy(), landmarks.numpy()

        # Filter bboxes by confidence score
        indices = np.where(scores > self.score_threshold)[0]
        bboxes = bboxes[indices]
        scores = scores[indices]
        landmarks = landmarks[indices]

        # Swap position of x, y coordinates
        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]

        bboxes = xyxy2xyxyn(bboxes, image.shape[0], image.shape[1])

        return bboxes, scores, landmarks

    @staticmethod
    def _preprocess(image: np.ndarray) -> tf.Tensor:
        """Processes input image

        Args:
            image (np.ndarray): image in numpy array

        Returns:
            image (np.ndarray): processed numpy array of image
        """
        image = image.astype(np.float32)
        image = tf.convert_to_tensor(image)

        return image



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2019 ZhongdaoWang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Utility functions for JDE model.

Modifications:
- Removed unused ratio, width and height padding return values in letterbox()
- Removed filtering detections by score_threshold in non_max_suppression since
    it's already performed in track.py before calling non_max_suppression
"""

from typing import List, Tuple

import cv2
import numpy as np
import torch
from torchvision.ops import nms

from peekingduck.pipeline.utils.bbox.transforms import xywh2xyxy


def decode_delta(delta: torch.Tensor, anchor_mesh: torch.Tensor) -> torch.Tensor:
    """Converts raw output to (x, y, w, h) format where (x, y) is the center,
    w is the width, and h is the height of the bounding box.

    Args:
        delta (torch.Tensor): Raw output from the YOLOLayer.
        anchor_mesh (torch.Tensor): Tensor containing the grid points and their
            respective anchor offsets.

    Returns:
        (torch.Tensor): Decoded bounding box tensor.
    """
    delta[..., :2] = delta[..., :2] * anchor_mesh[..., 2:] + anchor_mesh[..., :2]
    delta[..., 2:] = torch.exp(delta[..., 2:]) * anchor_mesh[..., 2:]
    return delta


def decode_delta_map(
    delta_map: torch.Tensor, anchors: torch.Tensor, device: torch.device
) -> torch.Tensor:
    """Decodes raw bounding box output in to (x, y, w, h) format where
    (x, y) is the center, w is the width, and h is the height.

    Args:
        delta_map (torch.Tensor): A tensor with the shape
            (batch_size, num_anchors, grid_height, grid_width, 4) containing
            raw bounding box predictions.
        anchors (torch.Tensor): A tensor with the shape (num_anchors, 4)
            containing the anchors used for the `YOLOLayer`.
        device (torch.device): The device which a `torch.Tensor` is on or
            will be allocated.

    Returns:
        (torch.Tensor): Tensor containing the decoded bounding boxes.
    """
    batch_size, num_anchors, grid_height, grid_width, _ = delta_map.shape
    anchor_mesh = generate_anchor(grid_height, grid_width, anchors, device)
    # Shape (num_anchors x grid_height x grid_width) x 4
    anchor_mesh = anchor_mesh.permute(0, 2, 3, 1).contiguous()
    anchor_mesh = anchor_mesh.unsqueeze(0).repeat(batch_size, 1, 1, 1, 1)
    pred_list = decode_delta(delta_map.view(-1, 4), anchor_mesh.view(-1, 4))
    pred_map = pred_list.view(batch_size, num_anchors, grid_height, grid_width, 4)
    return pred_map


def generate_anchor(
    grid_height: int, grid_width: int, anchor_wh: torch.Tensor, device: torch.device
) -> torch.Tensor:
    """Generates grid anchors for a single level.

    Args:
        grid_height (int): Height of feature map.
        grid_width (int): Width of feature map.
        anchor_wh (torch.Tensor): Width and height of the anchor boxes.
        device (torch.device): The device which a `torch.Tensor` is on or
            will be allocated.

    Returns:
        (torch.Tensor): Anchors of a feature map in a single level.
    """
    num_anchors = len(anchor_wh)
    y_vec, x_vec = torch.meshgrid(torch.arange(grid_height), torch.arange(grid_width))
    x_vec, y_vec = x_vec.to(device), y_vec.to(device)

    # Shape 2 x grid_height x grid_width
    mesh = torch.stack([x_vec, y_vec], dim=0)
    # Shape num_anchors x 2 x grid_height x grid_width
    mesh = mesh.unsqueeze(0).repeat(num_anchors, 1, 1, 1).float()
    # Shape num_anchors x 2 x grid_height x grid_width
    anchor_offset_mesh = (
        anchor_wh.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, grid_height, grid_width)
    )
    # Shape num_anchors x 4 x grid_height x grid_width
    anchor_mesh = torch.cat([mesh, anchor_offset_mesh], dim=1)
    return anchor_mesh


def letterbox(
    image: np.ndarray,
    height: int,
    width: int,
    color: Tuple[float, float, float] = (127.5, 127.5, 127.5),
) -> np.ndarray:
    """Resizes a rectangular image to a padded rectangular image.

    Args:
        image (np.ndarray): Image frame.
        height (int): Height of padded image.
        width (int): Width of padded image.
        color (Tuple[float, float, float]): Color used for padding around
            the image. (127.5, 127.5, 127.5) is chosen as it is used by the
            original project during model training.

    Returns:
        (np.ndarray): Padded rectangular image.
    """
    shape = image.shape[:2]  # shape = [height, width]
    ratio = min(float(height) / shape[0], float(width) / shape[1])
    # new_shape = [width, height]
    new_shape = (round(shape[1] * ratio), round(shape[0] * ratio))
    width_padding = (width - new_shape[0]) / 2
    height_padding = (height - new_shape[1]) / 2
    top = round(height_padding - 0.1)
    bottom = round(height_padding + 0.1)
    left = round(width_padding - 0.1)
    right = round(width_padding + 0.1)
    # resized, no border
    image = cv2.resize(image, new_shape, interpolation=cv2.INTER_AREA)
    # padded rectangular
    image = cv2.copyMakeBorder(
        image, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color
    )
    return image


def non_max_suppression(
    prediction: torch.Tensor, nms_threshold: float
) -> List[torch.Tensor]:
    """Removes detections with lower object confidence score than
    `score_threshold`. Non-Maximum Suppression to further filter detections.

    Args:
        prediction (torch.Tensor): Predicted bounding boxes.
        nms_threshold (float): Threshold for Intersection-over-Union values of
            the bounding boxes.

    Returns:
        (List[Optional[torch.Tensor]]): List of detections with shape
            (x1, y1, x2, y2, object_conf, class_score, class_pred). For
            detections which have all bounding boxes filtered by `nms`, the
            element will be `None` instead.
    """
    # Initializing this list with torch.empty will likely incur some additional
    # computational cost
    output: List[torch.Tensor] = [None for _ in range(len(prediction))]  # type: ignore
    for i, pred in enumerate(prediction):
        # From (center x, center y, width, height) to (x1, y1, x2, y2)
        pred[:, :4] = xywh2xyxy(pred[:, :4])
        # Non-maximum suppression
        nms_indices = nms(pred[:, :4], pred[:, 4], nms_threshold)
        det_max = pred[nms_indices]

        if len(det_max) == 0:  # pragma: no cover
            # This really shouldn't happen since nms will at worst leave one
            # bbox and suppress everything else
            continue
        # Add max detections to outputs
        output[i] = det_max if output[i] is None else torch.cat((output[i], det_max))

    return output


def scale_coords(
    img_size: List[int], coords: torch.Tensor, img0_size: Tuple[int, int]
) -> torch.Tensor:
    """Rescales bounding box coordinates (x1, y1, x2, y2) from `img_size` to
    `img0_size`.

    Args:
        img_size (List[int]): Model input size (w x h).
        coords (torch.Tensor): Bounding box coordinates.
        img0_size (Tuple[int, int]): Size of original video frame (h x w).

    Returns:
        (torch.Tensor): Bounding boxes with resized coordinates.
    """
    # gain = old / new
    gain = min(float(img_size[0]) / img0_size[1], float(img_size[1]) / img0_size[0])
    pad_x = (img_size[0] - img0_size[1] * gain) / 2  # width padding
    pad_y = (img_size[1] - img0_size[0] * gain) / 2  # height padding
    coords[:, [0, 2]] -= pad_x
    coords[:, [1, 3]] -= pad_y
    coords[:, :4] /= gain
    coords[:, :4] = torch.clamp(coords[:, :4], min=0)
    return coords



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Constants used for drawing
"""

# colors
CHAMPAGNE = (156, 223, 244)
BLIZZARD = (241, 232, 164)
VIOLET_BLUE = (188, 118, 119)
TOMATO = (77, 103, 255)
BROWN = (96, 109, 167)
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
PRIMARY_PALETTE_LENGTH = 5
PRIMARY_PALETTE = [CHAMPAGNE, BLIZZARD, TOMATO, VIOLET_BLUE, BROWN]

# constants for thickness
THIN = 1
THICK = 2
VERY_THICK = 3

# constant to fill shapes in cv2. To be replace line thickness
FILLED = -1

# constants for font scale
SMALL_FONTSCALE = 0.5
NORMAL_FONTSCALE = 1
BIG_FONTSCALE = 2

# constants used for image manipulation
LOWER_SATURATION = 0.5

# constants for pts
POINT_RADIUS = 5



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MoveNet model with model types: singlepose lightning/thunder, multipose lightning"""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.movenetv1.movenet_files.predictor import Predictor


class MoveNetModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """MoveNet model with model types: lightning, thunder."""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_valid_choice(
            "model_type",
            {"singlepose_lightning", "singlepose_thunder", "multipose_lightning"},
        )
        self.check_bounds(
            ["bbox_score_threshold", "keypoint_score_threshold"], "[0, 1]"
        )

        model_dir = self.download_weights()
        self.predictor = Predictor(
            model_dir,
            self.config["model_type"],
            self.weights["model_file"],
            self.config["resolution"],
            self.config["bbox_score_threshold"],
            self.config["keypoint_score_threshold"],
        )

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Predict poses from input frame

        Args:
            frame (np.array): image in numpy array

        Returns:
            Tuple of outputs of the model
            (bboxes, keypoints, keypoint_scores, keypoint_masks, keypoint_conns)

            bboxes (np.ndarray): Nx4 array of bboxes, N is number of detections
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates
            keypoints_scores (np.ndarray): Nx17 array of keypoint scores
            keypoints_conns (np.ndarray): NxD'x2 keypoint connections, where
                D' is the varying pair of valid keypoint connections per detection
        """
        return self.predictor.predict(frame)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to load the MTCNN model."""

import logging
from typing import Callable, List, Tuple

import tensorflow as tf

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def load_graph(file_path: str, inputs: List[str], outputs: List[str]) -> Callable:
    """Loads a frozen graph and wraps it into a function.

    Args:
        file_path (str): Path to the frozen graph or model.
        inputs (List[str]): List of input names.
        outputs (List[str]): List of tensor names to be returned.

    Returns:
        (Callable): A WrappedFunction which wraps a tf V1 piece of code in a
        function.
    """
    with tf.io.gfile.GFile(file_path, "rb") as graph_file:
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(graph_file.read())

        frozen_func = wrap_frozen_graph(graph_def, inputs, outputs)

        return frozen_func


def wrap_frozen_graph(
    graph_def: tf.compat.v1.GraphDef, inputs: List[str], outputs: List[str]
) -> Callable:
    """Wraps a frozen graph into a function.

    Args:
        graph_def (tf.compat.v1.GraphDef): A frozen graph in graph_def format.
        inputs (List[str]): List of input names.
        outputs (List[str]): List of tensor names to be returned.

    Return:
        (Callable): A wrapped_import function to perform your inference with.
    """

    def _imports_graph_def(
        img: tf.Tensor, min_size: tf.Tensor, factor: tf.Tensor, thresholds: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        prob, landmarks, box = tf.compat.v1.import_graph_def(
            graph_def,
            input_map={
                inputs[0]: img,
                inputs[1]: min_size,
                inputs[2]: thresholds,
                inputs[3]: factor,
            },
            return_elements=outputs,
            name="",
        )

        return box, prob, landmarks

    wrapped_import = tf.compat.v1.wrap_function(
        _imports_graph_def,
        [
            tf.TensorSpec(shape=[None, None, 3], dtype=tf.float32),
            tf.TensorSpec(shape=[], dtype=tf.float32),
            tf.TensorSpec(shape=[], dtype=tf.float32),
            tf.TensorSpec(shape=[3], dtype=tf.float32),
        ],
    )

    return wrapped_import



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Counts the number of detected boxes."""

from typing import Any, Dict

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Counts the total number of detected objects.

    Inputs:
        |bboxes_data|

    Outputs:
        |count_data|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Counts bboxes of object chosen in the frame.

        Note that this method requires that the bboxes returned to all belong
        to the same object category (for example, all "person").
        """
        return {"count": len(inputs["bboxes"])}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main class for MTCNN Model
"""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.base import (
    ThresholdCheckerMixin,
    WeightsDownloaderMixin,
)
from peekingduck.pipeline.nodes.model.mtcnnv1.mtcnn_files.detector import Detector


class MTCNNModel(ThresholdCheckerMixin, WeightsDownloaderMixin):
    """MTCNN model to detect face bboxes and landmarks."""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.check_bounds("min_size", "(0, +inf]")
        self.check_bounds(
            ["network_thresholds", "scale_factor", "score_threshold"], "[0, 1]"
        )

        model_dir = self.download_weights()
        self.detector = Detector(
            model_dir,
            self.config["model_type"],
            self.weights["model_file"],
            self.config["model_nodes"],
            self.config["min_size"],
            self.config["scale_factor"],
            list(map(float, self.config["network_thresholds"])),
            self.config["score_threshold"],
        )

    def predict(self, frame: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, scores and landmarks

        Args:
            frame (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): numpy array of detected bboxes
            scores (np.ndarray): numpy array of confidence scores
            landmarks (np.ndarray): numpy array of facial landmarks
        """
        assert isinstance(frame, np.ndarray)

        return self.detector.predict_object_bbox_from_image(frame)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Converts bounding boxes to a single point of reference."""

from typing import Any, Dict, Tuple

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Converts bounding boxes to a single point which is the bottom midpoint
    of the bounding box.

    This node is primarily used for zone counting. The bottom midpoint is an
    unambiguous way of telling whether an object is in the zone specified, as
    the bottom midpoint usually corresponds to the point where the object is
    located.

    Inputs:
        |img_data|

        |bboxes_data|

    Outputs:
        |btm_midpoint_data|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.img_size = None

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Converts bounding boxes to a single point of reference for use in
        zone analytics.
        """
        # get xy midpoint of each bbox (x1, y1, x2, y2)
        # This is calculated as x is (x1-x2)/2 and y is y2
        bboxes = inputs["bboxes"]
        frame = inputs["img"]
        self.img_size = (frame.shape[1], frame.shape[0])  # type:ignore
        return {
            "btm_midpoint": [
                self._xy_on_img(((bbox[0] + bbox[2]) / 2), bbox[3]) for bbox in bboxes
            ]
        }

    def _xy_on_img(self, pt_x: float, pt_y: float) -> Tuple[int, int]:
        """Return the int x y points of the midpoint on the original image"""
        assert self.img_size is not None
        return (int(pt_x * self.img_size[0]), int(pt_y * self.img_size[1]))



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Preprocessing functions for PoseNet
"""

from typing import Tuple

import cv2
import numpy as np

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.constants import (
    IMAGE_NET_MEAN,
)


def rescale_image(
    source_img: np.ndarray,
    input_res: Tuple[int, int],
    scale_factor: float = 1.0,
    output_stride: int = 16,
    model_type: str = "mobilenet",
) -> Tuple[np.ndarray, np.ndarray]:
    """Rescale the image by a scale factor while ensuring it has a valid output
    stride

    Args:
        source_img (np.array): image for inference
        input_res (Tuple[int, int]): input height and width of image
        scale_factor (float): ratio to scale image
        output_stride (int): output stride to convert output indices to image coordinates
        model_type (str): specified model type (refer to modelconfig.yml)

    Returns:
        image_processed (np.array): processed image array
        scale (np.array): factor to scale height and width
    """
    target_width, target_height = _get_valid_resolution(
        input_res[0] * scale_factor,
        input_res[1] * scale_factor,
        output_stride=output_stride,
    )

    scale = np.array(
        [source_img.shape[1] / target_width, source_img.shape[0] / target_height]
    )

    image_processed = _rescale_image(
        source_img, target_width, target_height, model_type
    )

    return image_processed, scale


def _get_valid_resolution(
    width: float, height: float, output_stride: int = 16
) -> Tuple[int, int]:
    """Get valid height and width divisible by output stride"""
    target_width = (int(width) // output_stride) * output_stride + 1
    target_height = (int(height) // output_stride) * output_stride + 1
    return target_width, target_height


def _rescale_image(
    source_img: np.ndarray, target_width: int, target_height: int, model_type: str
) -> np.ndarray:
    """Apply different preprocessing depending on model type - mobilenet or resnet

    For mobilenet version, RGB values were scaled (2.0 / 255.0) - 1.0
    For resnet version, IMAGE_NET_MEAN were added to the RGB values
    """
    image_processed = cv2.resize(
        source_img, (target_width, target_height), interpolation=cv2.INTER_LINEAR
    )
    # Color conversion part of original repo code
    image_processed = cv2.cvtColor(image_processed, cv2.COLOR_BGR2RGB).astype(
        np.float32
    )

    if model_type == "resnet":
        image_processed += IMAGE_NET_MEAN
    else:
        image_processed = image_processed * (2.0 / 255.0) - 1.0

    image_processed = image_processed.reshape(1, target_height, target_width, 3)
    return image_processed



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Blurs area bounded by bounding boxes over detected object.
"""

from typing import Any, Dict, List

import cv2
import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Blurs area bounded by bounding boxes on image.

    The :mod:`draw.blur_bbox` node blurs the areas of the image bounded by the
    bounding boxes output from an object detection model.

    Inputs:
        |img_data|

        |bboxes_data|

    Outputs:
        |img_data|

    Configs:
        blur_kernel_size (:obj:`int`): **default = 50**. |br|
            This defines the kernel size used in the blur filter. Larger values
            of ``blur_kernel_size`` gives more intense blurring.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.blur_kernel_size = self.config["blur_kernel_size"]

    @staticmethod
    def _blur(
        bboxes: List[np.ndarray], image: np.ndarray, blur_kernel_size: int
    ) -> np.ndarray:
        """Blurs the area bounded by bbox in an image."""
        height = image.shape[0]
        width = image.shape[1]

        for bbox in bboxes:
            x_1, y_1, x_2, y_2 = bbox
            y_1, y_2 = int(y_1 * height), int(y_2 * height)
            x_1, x_2 = int(x_1 * width), int(x_2 * width)

            # to get the area bounded by bbox
            bbox_image = image[y_1:y_2, x_1:x_2, :]

            # apply the blur using blur filter from opencv
            blur_bbox_image = cv2.blur(bbox_image, (blur_kernel_size, blur_kernel_size))
            image[y_1:y_2, x_1:x_2, :] = blur_bbox_image

        return image

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads the image input and returns the image, with the areas bounded
        by the bboxes blurred.

        Args:
            inputs (dict): Dictionary of inputs with keys "img", "bboxes"

        Returns:
            outputs (dict): Output in dictionary format with key "img"
        """
        blurred_img = self._blur(inputs["bboxes"], inputs["img"], self.blur_kernel_size)
        outputs = {"img": blurred_img}
        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🕺 Fast Pose Estimation model."""

from typing import Any, Dict

import cv2
import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.model.movenetv1 import movenet_model


class Node(AbstractNode):
    """MoveNet node that initializes a MoveNet model to detect human poses from
    an image.

    The MoveNet node is capable of detecting up to 6 human figures for
    multipose lightning and single person for singlepose lightning/thunder. If
    there are more than 6 persons in the image, multipose lightning will only
    detect 6. This also applies to singlepose models, where only 1 person will
    be detected in a multi persons image, do take note that detection
    performance will suffer when using singlepose models on multi persons
    images. 17 keypoints are estimated and the keypoint indices table can be
    found :ref:`here <whole-body-keypoint-ids>`.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |keypoints_data|

        |keypoint_scores_data|

        |keypoint_conns_data|

        |bbox_labels_data|

    Configs:
        model_type (:obj:`str`):
            **{"
            singlepose_lightning", "singlepose_thunder", "multipose_lightning"
            },  default="multipose_lightning"** |br|
            Defines the detection model for MoveNet either single or multi pose.
            Lightning is smaller and faster but less accurate than Thunder
            version.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        resolution (:obj:`Dict`): **default = { height: 256, width: 256 }** |br|
            Dictionary of resolutions of input array to different MoveNet models.
            Only multipose allows dynamic shape in multiples of 32 (recommended
            256). Default will be the resolution for multipose lightning model.
        bbox_score_threshold (:obj:`float`): **[0,1], default = 0.2** |br|
            Detected bounding box confidence score threshold, only boxes above
            threshold will be kept in the output.
        keypoint_score_threshold (:obj:`float`): **[0,1], default = 0.3** |br|
            Detected keypoints confidence score threshold, only keypoints above
            threshold will be kept in output.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = movenet_model.MoveNetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Function that reads the image input and returns the bboxes,
        keypoints, keypoints confidence scores, keypoint connections and
        bounding box labels of the persons detected.

        Args:
            inputs (Dict[str, Any]): Dictionary of inputs with key "img".

        Returns:
            (Dict[str, Any]): bbox output in dictionary format with keys
            "bboxes", "keypoints", "keypoint_scores", "keypoint_conns", and
            "bbox_labels".
        """
        image = cv2.cvtColor(inputs["img"], cv2.COLOR_BGR2RGB)
        bboxes, keypoints, keypoint_scores, keypoint_conns = self.model.predict(image)
        bbox_labels = np.array(["person"] * len(bboxes))
        bboxes = np.clip(bboxes, 0, 1)

        return {
            "bboxes": bboxes,
            "bbox_labels": bbox_labels,
            "keypoints": keypoints,
            "keypoint_conns": keypoint_conns,
            "keypoint_scores": keypoint_scores,
        }



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🕺 High-Resolution Network: Deep high-resolution representation learning for
human pose estimation. Requires an object detector.
"""


from typing import Any, Dict

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.model.hrnetv1 import hrnet_model


class Node(AbstractNode):
    """Initializes and uses HRNet model to infer poses from detected bboxes.
    Note that HRNet must be used in conjunction with an object detector applied
    prior.

    The HRNet applied to human pose estimation uses the representation head,
    called HRNetV1.

    The HRNet node is capable of detecting single human figures simultaneously
    per inference, with 17 keypoints estimated for each detected human figure.
    The keypoint indices table can be found
    :ref:`here <whole-body-keypoint-ids>`.

    Inputs:
        |img_data|

        |bboxes_data|

    Outputs:
        |keypoints_data|

        |keypoint_scores_data|

        |keypoint_conns_data|

    Configs:
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        resolution (:obj:`Dict`):
            **default = { height: 192, width: 256 }**. |br|
            Resolution of input array to HRNet model.
        score_threshold (:obj:`float`): **[0, 1], default = 0.1**. |br|
            Threshold to determine if detection should be returned

    References:
        Deep High-Resolution Representation Learning for Visual Recognition:
        https://arxiv.org/abs/1908.07919
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = hrnet_model.HRNetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads the bbox input and returns the poses and pose bbox of the
        specified objects chosen to be detected.
        """
        keypoints, keypoint_scores, keypoint_conns = self.model.predict(
            inputs["img"], inputs["bboxes"]
        )

        outputs = {
            "keypoints": keypoints,
            "keypoint_scores": keypoint_scores,
            "keypoint_conns": keypoint_conns,
        }
        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for Tensorflow Graphed models
"""

import logging
import os
from typing import Callable, List

import tensorflow as tf

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def wrap_frozen_graph(
    graph_def: tf.compat.v1.GraphDef, inputs: List[str], outputs: List[str]
) -> Callable:
    """
    Wraps the graph into a function. This is akin to a model.predict() function
    in keras. When doing inference, simply do frozen_function(tf.cast(x, float))[0].
    It will return your predicted values.

    Args:
        graph_def (tf.compat.v1.GraphDef): The frozen graph in graph_def format
        inputs (List[str]): The name(s) of the input nodes from your graph,
            e.g., ['inputs']
        outputs (List[str]): The name(s) of your output nodes from your graph,
            e.g., ['heatmap', 'offsets', 'displacement_fwd',
            'displacement_bwd']

    Returns:
        (Callable): A wrapped_import function to perform your inference with.
    """

    def _imports_graph_def() -> None:  # this needs to be here because of graph_def
        tf.compat.v1.import_graph_def(graph_def, name="")

    wrapped_import = tf.compat.v1.wrap_function(_imports_graph_def, [])
    import_graph = wrapped_import.graph

    return wrapped_import.prune(
        tf.nest.map_structure(import_graph.as_graph_element, inputs),
        tf.nest.map_structure(import_graph.as_graph_element, outputs),
    )


def load_graph(file_path: str, inputs: List[str], outputs: List[str]) -> Callable:
    """Loads the graph."""
    with tf.io.gfile.GFile(file_path, "rb") as graph_file:
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(graph_file.read())

        frozen_func = wrap_frozen_graph(graph_def, inputs, outputs)

        return frozen_func


def print_inputs(graph_def: tf.compat.v1.GraphDef) -> None:
    """Prints the input nodes of graph_def."""
    with tf.Graph().as_default() as graph:  # pylint: disable=not-context-manager
        tf.import_graph_def(graph_def, name="")

    input_list = []
    for graph_op in graph.get_operations():  # tensorflow.python.framework.ops.Operation
        if graph_op.type == "Placeholder":
            input_list.append(graph_op.name)

        logger.info(f"Inputs: {input_list}")


def print_outputs(graph_def: tf.compat.v1.GraphDef) -> None:
    """Prints the output nodes of graph_def."""
    name_list = []
    input_list = []
    for node in graph_def.node:  # tensorflow.core.framework.node_def_pb2.NodeDef
        name_list.append(node.name)
        input_list.extend(node.input)

    outputs = set(name_list) - set(input_list)
    logger.info(f"Outputs: {list(outputs)}")



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions which convert bbox coordinates from one format to
another.
"""

import numpy as np
import torch


def tlwh2xyah(inputs: np.ndarray) -> np.ndarray:
    """Converts from [t, l, w, h] to [x, y, a, h] format.

    (t, l) is the coordinates of the top left corner, w is the width, and h is
    the height. (x, y) is the coordinates of the object center, `a` is the
    aspect ratio, and `h` is the height. Aspect ratio is `width / height`.

    [x, y, a, h] is calculated as:
    x = (t + w) / 2
    y = (l + h) / 2
    a = w / h
    h = h

    Example:
        >>> a = tlwh2xyah(inputs=np.array([1.0, 2.0, 30.0, 40.0]))
        >>> a
        array([16.0, 22.0, 0.75, 40.0])

    Args:
        inputs (np.ndarray): Input bounding box (1-d array) with the format
            `(top left x, top left y, width, height)`.

    Returns:
        (np.ndarray): Bounding box with the format `(center x, center y, aspect
        ratio,height)`.
    """
    outputs = np.asarray(inputs).copy()
    outputs[:2] += outputs[2:] / 2
    outputs[2] /= outputs[3]
    return outputs


def tlwh2xyxyn(inputs: np.ndarray, height: int, width: int) -> np.ndarray:
    """Converts from [t, l, w, h] to normalized [x1, y1, x2, y2] format.
    Normalized coordinates are w.r.t. original image size.

    (t, l) is the coordinates of the top left corner, w is the width, and h is
    the height. (x1, y1) and (x2, y2) are the normalized coordinates of top
    left and bottom right, respectively.

    [x1, y1, x2, y2] is calculated as:
    x1 = t / width
    y1 = l / height
    x2 = (t + w) / width
    y2 = (l + h) / height

    Example:
        >>> a = tlwh2xyxyn(inputs=np.array([[1.0, 2.0, 30.0, 40.0]]), height=100, width=200)
        >>> a
        array([[0.005, 0.02, 0.155, 0.42]])

    Args:
        inputs (np.ndarray): Input bounding boxes (2-d array) each with the
            format `(top left x, top left y, width, height)`.
        height (int): Height of the image frame.
        height (int): Width of the image frame.

    Returns:
        (np.ndarray): Bounding boxes with the format `normalized (top left x,
        top left y, bottom right x, bottom right y)`.
    """
    outputs = np.empty_like(inputs)
    outputs[:, 0] = inputs[:, 0] / width
    outputs[:, 1] = inputs[:, 1] / height
    outputs[:, 2] = (inputs[:, 0] + inputs[:, 2]) / width
    outputs[:, 3] = (inputs[:, 1] + inputs[:, 3]) / height
    return outputs


def xywh2xyxy(inputs: torch.Tensor) -> torch.Tensor:
    """Converts from [x, y, w, h] to [x1, y1, x2, y2] format.

    (x, y) is the object center, w is the width, and h is the height. (x1, y1)
    is the top left corner and (x2, y2) is the bottom right corner.

    [x1, y1, x2, y2] is calculated as:
    x1 = x - w / 2
    y1 = y - h / 2
    x2 = x + w / 2
    y2 = y + w / 2

    Example:
        >>> a = xywh2xyxy(inputs=torch.Tensor([[100, 200, 30, 40]]))
        >>> a
        tensor([[ 85.0, 180.0, 115.0, 220.0]])

    Args:
        inputs (torch.Tensor): Input bounding boxes (2-d array) each with the
            format `(center x, center y, width, height)`.

    Returns:
        (torch.Tensor): Bounding boxes with the format `(top left x, top left y,
        bottom right x, bottom right y)`.
    """
    outputs = torch.empty_like(inputs)
    outputs[:, 0] = inputs[:, 0] - inputs[:, 2] / 2
    outputs[:, 1] = inputs[:, 1] - inputs[:, 3] / 2
    outputs[:, 2] = inputs[:, 0] + inputs[:, 2] / 2
    outputs[:, 3] = inputs[:, 1] + inputs[:, 3] / 2

    return outputs


def xyxy2tlwh(inputs: np.ndarray) -> np.ndarray:
    """Converts from [x1, y1, x2, y2] to [t, l, w, h].

    (x1, y1) and (x2, y2) are the coordinates of top left and bottom right,
    respectively. (t, l) is the coordinates of the top left corner, w is the
    width, and h is the height.

    [t, l, w, h] is calculated as:
    t = x1
    l = y1
    w = x2 - x1
    h = y2 - y1

    Example:
        >>> a = xyxy2tlwh(inputs=np.array([1.0, 2.0, 30.0, 40.0]))
        >>> a
        array([1, 2, 29, 38])

    Args:
        inputs (np.ndarray): Input bounding box (1-d array) each with the
            format `(top left x, top left y, bottom right x, bottom right y)`.

    Returns:
        (np.ndarray): Bounding box with the format `(top left x, top left y,
        width, height)`.
    """
    outputs = np.asarray(inputs).copy()
    outputs[2:] -= outputs[:2]
    return outputs


def xyxy2xyxyn(inputs: np.ndarray, height: float, width: float) -> np.ndarray:
    """Converts from [x1, y1, x2, y2] to normalized [x1, y1, x2, y2].
    Normalized coordinates are w.r.t. original image size.

    (x1, y1) is the top left corner and (x2, y2) is the bottom right corner.

    Normalized [x1, y1, x2, y2] is calculated as:
    Normalized x1 = x1 / width
    Normalized y1 = y1 / height
    Normalized x2 = x2 / width
    Normalized y2 = y2 / height

    Example:
        >>> a = xyxy2xyxyn(inputs=np.array([[1.0, 2.0, 30.0, 40.0]]), height=100, width=200)
        >>> a
        array([[0.005, 0.02, 0.15, 0.4]])

    Args:
        inputs (np.ndarray): Input bounding boxes (2-d array) each with the
            format `(top left x, top left y, bottom right x, bottom right y)`.

    Returns:
        (np.ndarray): Bounding boxes with the format `normalized (top left x,
        top left y, bottom right x, bottom right y)`.
    """
    outputs = np.empty_like(inputs)
    outputs[:, [0, 2]] = inputs[:, [0, 2]] / width
    outputs[:, [1, 3]] = inputs[:, [1, 3]] / height

    return outputs


def xyxyn2tlwh(inputs: np.ndarray, height: float, width: float) -> np.ndarray:
    """Converts from normalized [x1, y1, x2, y2] to [t, l, w, h] format.
    Normalized coordinates are w.r.t. original image size.

    (t, l) is the coordinates of the top left corner, w is the width, and h is
    the height. (x1, y1) and (x2, y2) are the normalized coordinates of top
    left and bottom right, respectively.

    [t, l, w, h] is calculated as:
    t = x1 * width
    l = y1 * height
    w = (x2 - x1) * width
    h = (y2 - y1) * height

    Example:
        >>> a = xyxyn2tlwh(inputs=np.array([[0.0, 0.02, 0.3, 0.4]]), height=100, width=200)
        >>> a
        array([[ 0.0  2.0 60.0 38.0]])

    Args:
        inputs (np.ndarray): Input bounding boxes (2-d array) each with the
            format `normalized (top left x, top left y, bottom right x, bottom
            right y)`.
        height (int): Height of the image frame.
        height (int): Width of the image frame.

    Returns:
        (np.ndarray): Bounding boxes with the format `(top left x, top left y,
        width, height)`.
    """
    outputs = np.empty_like(inputs)
    outputs[:, 0] = inputs[:, 0] * width
    outputs[:, 1] = inputs[:, 1] * height
    outputs[:, 2] = (inputs[:, 2] - inputs[:, 0]) * width
    outputs[:, 3] = (inputs[:, 3] - inputs[:, 1]) * height

    return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws keypoints on a detected pose.
"""

from typing import Any, Dict

from peekingduck.pipeline.nodes.draw.utils.pose import draw_human_poses
from peekingduck.pipeline.nodes.abstract_node import AbstractNode


class Node(AbstractNode):
    """Draws poses onto image.

    The :mod:`draw.poses` node uses the :term:`keypoints`,
    :term:`keypoint_scores`, and :term:`keypoint_conns` predictions from pose
    models to draw the human poses onto the image. For better understanding,
    check out the pose models such as :mod:`HRNet <model.hrnet>` and
    :mod:`PoseNet <model.posenet>`.

    Inputs:
        |img_data|

        |keypoints_data|

        |keypoint_scores_data|

        |keypoint_conns_data|

    Outputs:
        |none_output_data|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.keypoint_dot_color = tuple(self.config["keypoint_dot_color"])
        self.keypoint_connect_color = tuple(self.config["keypoint_connect_color"])
        self.keypoint_text_color = tuple(self.config["keypoint_text_color"])

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws pose details onto input image.

        Args:
            inputs (dict): Dictionary with keys "img", "keypoints", and
                "keypoint_conns".
        """
        draw_human_poses(inputs["img"], inputs["keypoints"], inputs["keypoint_conns"])
        return {}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Records the nodes' outputs to a CSV file."""

import logging
import textwrap
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.output.utils.csvlogger import CSVLogger


class Node(AbstractNode):
    """Tracks user-specified parameters and outputs the results in a CSV file.

    Inputs:
        ``all`` (:obj:`List`): A placeholder that represents a flexible input.
        Actual inputs to be written into the CSV file can be configured in
        ``stats_to_track``.

    Outputs:
        |none_output_data|

    Configs:
        stats_to_track (:obj:`List`):
            **default = ["keypoints", "bboxes", "bbox_labels"]**. |br|
            Parameters to log into the CSV file. The chosen parameters must be
            present in the data pool.
        file_path (:obj:`str`):
            **default = "PeekingDuck/data/stats.csv"**. |br|
            Path of the CSV file to be saved. The resulting file name would have an appended
            timestamp.
        logging_interval (:obj:`int`): **default = 1**. |br|
            Interval between each log, in terms of seconds.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.logger = logging.getLogger(__name__)
        self.logging_interval = int(self.logging_interval)  # type: ignore
        self.file_path = Path(self.file_path)  # type: ignore
        # check if file_path has a '.csv' extension
        if self.file_path.suffix != ".csv":
            raise ValueError("Filepath must have a '.csv' extension.")

        self._file_path_datetime = self._append_datetime_file_path(self.file_path)
        self._stats_checked = False
        self.stats_to_track: List[str]
        self.csv_logger = CSVLogger(
            self._file_path_datetime, self.stats_to_track, self.logging_interval
        )

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Writes the current state of the tracked statistics into
        the csv file as a row entry

        Args:
            inputs (dict): The data pool of the pipeline.

        Returns:
            outputs: [None]
        """
        # reset and terminate when there are no more data
        if inputs["pipeline_end"]:
            self._reset()
            return {}

        if not self._stats_checked:
            self._check_tracked_stats(inputs)
            # self._stats_to_track might change after the check
            self.csv_logger = CSVLogger(
                self._file_path_datetime, self.stats_to_track, self.logging_interval
            )

        self.csv_logger.write(inputs, self.stats_to_track)

        return {}

    def _check_tracked_stats(self, inputs: Dict[str, Any]) -> None:
        """Checks whether user input statistics is present in the data pool
        of the pipeline. Statistics not present in data pool will be
        ignored and dropped.
        """
        valid = []
        invalid = []

        for stat in self.stats_to_track:
            if stat in inputs:
                valid.append(stat)
            else:
                invalid.append(stat)

        if invalid:
            msg = textwrap.dedent(
                f"""\
                {invalid} are not valid outputs.
                Data pool only has this outputs: {list(inputs.keys())}
                Only {valid} will be logged in the csv file.
                """
            )
            self.logger.warning(msg)

        # update stats_to_track with valid stats found in data pool
        self.stats_to_track = valid
        self._stats_checked = True

    def _reset(self) -> None:
        del self.csv_logger

        # initialize for use in run
        self._stats_checked = False

    @staticmethod
    def _append_datetime_file_path(file_path: Path) -> Path:
        """Append time stamp to the filename."""
        current_time = datetime.now()
        # output as '240621-15-09-13'
        time_str = current_time.strftime("%d%m%y-%H-%M-%S")

        # append timestamp to filename before extension
        # Format: filename_timestamp.extension
        file_path_with_timestamp = file_path.with_name(
            f"{file_path.stem}_{time_str}{file_path.suffix}"
        )

        return file_path_with_timestamp



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""FairMOT Multi-object Tracker.

Modifications include:
- Rearranged comments to they appear before the relevant lines of code
- Refactor variable names in update() for clarity
- Refactor subtract_stracks() to use list comprehension
- Refactor combine_stracks() to use bool for dictionary values instead
- Renamed head keys from hm, wh, and reg to heatmap, size, and offset
    respectively
- Refactor model prediction to a separate method
"""

import logging
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import torch
import torch.nn.functional as F

from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files import matching
from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.decoder import Decoder
from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.dla import DLASeg
from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.kalman_filter import (
    KalmanFilter,
)
from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.track import (
    STrack,
    TrackState,
)
from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.utils import (
    letterbox,
    transpose_and_gather_feat,
)
from peekingduck.pipeline.utils.bbox.transforms import tlwh2xyxyn, xyxy2tlwh


class Tracker:  # pylint: disable=too-many-instance-attributes
    """FairMOT Multi-object Tracker.

    Args:
        config (Dict[str, Any]): Model configuration options.
        model_dir (Path): Directory to model weights files.
        frame_rate (float): Frame rate of the current video sequence, used
            for computing size of track buffer.
    """

    heads = {"hm": 1, "wh": 4, "id": 128, "reg": 2}
    down_ratio = 4

    mean = np.array([0.408, 0.447, 0.470], dtype=np.float32).reshape((1, 1, 3))
    std = np.array([0.289, 0.274, 0.278], dtype=np.float32).reshape((1, 1, 3))

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        frame_rate: float,
        model_type: str,
        model_file: Dict[str, str],
        input_size: List[int],
        max_per_image: int,
        min_box_area: int,
        track_buffer: int,
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.model_type = model_type
        self.model_path = model_dir / model_file[self.model_type]
        self.input_size = input_size
        self.max_per_image = max_per_image
        self.min_box_area = min_box_area
        self.track_buffer = track_buffer
        self.score_threshold = score_threshold

        self.model = self._create_model()

        self.tracked_stracks: List[STrack] = []
        self.lost_stracks: List[STrack] = []
        self.removed_stracks: List[STrack] = []

        self.frame_id = 0
        self.max_time_lost = int(frame_rate / 30.0 * self.track_buffer)

        self.decoder = Decoder(self.max_per_image, self.down_ratio)
        self.kalman_filter = KalmanFilter()

    @torch.no_grad()
    def predict(
        self, padded_image: torch.Tensor, image: np.ndarray
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Predicts bounding boxes from the image and their associated Re-ID
        embeddings.

        Args:
            padded_image (torch.Tensor): Preprocessed image with letterbox
                resizing and colour normalisation.
            image (np.ndarray): The original video frame.

        Returns:
            (Tuple[torch.Tensor, torch.Tensor]): The predicted bounding boxes
            and their associated Re-ID embeddings.
        """
        output = self.model(padded_image)
        heatmap = output["hm"].sigmoid_()
        size = output["wh"]
        id_feature = F.normalize(output["id"], dim=1)
        offset = output["reg"]

        detections, indices = self.decoder(
            heatmap, size, offset, image.shape, padded_image.shape
        )
        id_feature = transpose_and_gather_feat(id_feature, indices)
        id_feature = id_feature.squeeze(0).cpu().numpy()

        id_feature = id_feature[detections[:, 4] > self.score_threshold]
        return detections, id_feature

    def track_objects_from_image(
        self, image: np.ndarray
    ) -> Tuple[List[np.ndarray], List[int], List[float]]:
        """Tracks detections from the current video frame.

        Args:
            image (np.ndarray): The current video frame.

        Returns:
            (Tuple[List[np.ndarray], List[str], List[float]]): A tuple of
            - Numpy array of detected bounding boxes.
            - List of track IDs.
            - List of detection confidence scores.
        """
        image_size = image.shape[:2]
        padded_image = self._preprocess(image)
        padded_image = torch.from_numpy(padded_image).to(self.device).unsqueeze(0)

        detections, embeddings = self.predict(padded_image, image)
        online_targets = self.update(detections, embeddings)
        online_tlwhs = []
        online_ids = []
        scores = []
        for target in online_targets:
            tlwh = target.tlwh
            vertical = tlwh[2] / tlwh[3] > 1.6
            if not vertical and tlwh[2] * tlwh[3] > self.min_box_area:
                online_tlwhs.append(tlwh)
                online_ids.append(target.track_id)
                scores.append(target.score.item())
        if not online_tlwhs:
            return online_tlwhs, online_ids, scores

        bboxes = self._postprocess(np.asarray(online_tlwhs), image_size)

        return bboxes, online_ids, scores

    def update(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        self, pred_detections: torch.Tensor, pred_embeddings: torch.Tensor
    ) -> List[STrack]:
        """Associates the detections with corresponding tracklets and also
        handles lost, removed, re-found and active tracklets.

        Args:
            pred_detections (torch.Tensor): Detections from the image, has the
                shape [N, 5].
            pred_embeddings (torch.Tensor): Re-ID embedding corresponding to
                each detection, has the shape [N, 128].
        Returns:
            (List[STrack]): The list contains information regarding the
            online tracklets for the received image tensor.
        """
        self.frame_id += 1
        activated_stracks = []
        refind_stracks = []
        lost_stracks = []
        removed_stracks = []

        # Step 1: Network forward, get detections & embeddings
        if len(pred_detections) > 0 and len(pred_embeddings) > 0:
            # Detections is list of (x1, y1, x2, y2, object_conf, class_score,
            # class_pred) class_pred is the embeddings.
            detections = [
                STrack(xyxy2tlwh(xyxys[:4]), xyxys[4], emb, 30)
                for (xyxys, emb) in zip(pred_detections[:, :5], pred_embeddings)
            ]
        else:
            detections = []

        # Add newly detected tracklets to tracked_stracks
        unconfirmed = []
        tracked_stracks: List[STrack] = []
        for track in self.tracked_stracks:
            if track.is_activated:
                # Active tracks are added to the local list 'tracked_stracks'
                tracked_stracks.append(track)
            else:
                # previous tracks which are not active in the current frame
                # are added in unconfirmed list
                unconfirmed.append(track)

        # Step 2: First association, with embedding
        # Combining currently tracked_stracks and lost_stracks
        strack_pool = _combine_stracks(tracked_stracks, self.lost_stracks)
        # Predict the current location with KF
        STrack.multi_predict(strack_pool)

        # The dists is a matrix of distances of the detection with the tracks
        # in strack_pool
        dists = matching.embedding_distance(strack_pool, detections)
        dists = matching.fuse_motion(self.kalman_filter, dists, strack_pool, detections)
        # matches is the array for corresponding matches of the detection
        # with the corresponding strack_pool
        (
            matches,
            unmatched_track_indices,
            unmatched_det_indices,
        ) = matching.linear_assignment(dists, threshold=0.4)

        for tracked_idx, det_idx in matches:
            # tracked_idx is the id of the track and det_idx is the detection
            track = strack_pool[tracked_idx]
            det = detections[det_idx]
            if track.state == TrackState.TRACKED:
                # If the track is active, add the detection to the track
                track.update(detections[det_idx], self.frame_id)
                activated_stracks.append(track)
            else:
                # We have obtained a detection from a track which is not
                # active, hence put the track in refind_stracks list
                track.re_activate(det, self.frame_id)
                refind_stracks.append(track)

        # None of the steps below happen if there are no undetected tracks.
        # Step 3: Second association, with IOU
        # detections is now a list of the unmatched detections
        detections = [detections[i] for i in unmatched_det_indices]
        # This is container for stracks which were tracked till the
        # previous frame but no detection was found for it in the current frame
        r_tracked_stracks = []
        for i in unmatched_track_indices:
            if strack_pool[i].state == TrackState.TRACKED:
                r_tracked_stracks.append(strack_pool[i])
        dists = matching.iou_distance(r_tracked_stracks, detections)
        # matches is the list of detections which matched with corresponding
        # tracks by IOU distance method
        (
            matches,
            unmatched_track_indices,
            unmatched_det_indices,
        ) = matching.linear_assignment(dists, threshold=0.5)
        # Same process done for some unmatched detections, but now considering
        # IOU_distance as measure
        for tracked_idx, det_idx in matches:
            track = r_tracked_stracks[tracked_idx]
            det = detections[det_idx]
            if track.state == TrackState.TRACKED:
                track.update(det, self.frame_id)
                activated_stracks.append(track)
            else:  # pragma: no cover
                # This shouldn't be reached, r_tracked_stracks only takes in
                # tracks with TrackState.TRACKED from above
                track.re_activate(det, self.frame_id)
                refind_stracks.append(track)
        # If no detections are obtained for tracks (unmatched_track_indices),
        # the tracks are added to lost_tracks and are marked lost
        for i in unmatched_track_indices:
            track = r_tracked_stracks[i]
            if track.state != TrackState.LOST:
                track.mark_lost()
                lost_stracks.append(track)

        # Deal with unconfirmed tracks, usually tracks with only one beginning
        # frame
        detections = [detections[i] for i in unmatched_det_indices]
        dists = matching.iou_distance(unconfirmed, detections)
        (
            matches,
            unconfirmed_track_indices,
            unmatched_det_indices,
        ) = matching.linear_assignment(dists, threshold=0.7)
        for tracked_idx, det_idx in matches:
            unconfirmed[tracked_idx].update(detections[det_idx], self.frame_id)
            activated_stracks.append(unconfirmed[tracked_idx])
        # The tracks which are yet not matched
        for i in unconfirmed_track_indices:
            track = unconfirmed[i]
            track.mark_removed()
            removed_stracks.append(track)

        # after all these confirmation steps, if a new detection is found, it
        # is initialized for a new track
        # Step 4: Init new stracks
        for i in unmatched_det_indices:
            track = detections[i]
            if track.score < self.score_threshold:  # pragma: no cover
                # This shouldn't be reached since we already rejected proposals
                # on basis of object confidence score in predict()
                continue
            track.activate(self.kalman_filter, self.frame_id)
            activated_stracks.append(track)

        # Step 5: Update state
        # If the tracks are lost for more frames than the threshold number, the
        # tracks are removed.
        for track in self.lost_stracks:
            if self.frame_id - track.end_frame > self.max_time_lost:
                track.mark_removed()
                removed_stracks.append(track)

        # Update the self.tracked_stracks and self.lost_stracks using the
        # updates in this step.
        self.tracked_stracks = [
            t for t in self.tracked_stracks if t.state == TrackState.TRACKED
        ]
        self.tracked_stracks = _combine_stracks(self.tracked_stracks, activated_stracks)
        self.tracked_stracks = _combine_stracks(self.tracked_stracks, refind_stracks)
        self.lost_stracks = _subtract_stracks(self.lost_stracks, self.tracked_stracks)
        self.lost_stracks.extend(lost_stracks)
        self.lost_stracks = _subtract_stracks(self.lost_stracks, self.removed_stracks)
        self.removed_stracks.extend(removed_stracks)
        self.tracked_stracks, self.lost_stracks = _remove_duplicate_stracks(
            self.tracked_stracks, self.lost_stracks
        )

        # get scores of lost tracks
        output_stracks = [track for track in self.tracked_stracks if track.is_activated]

        return output_stracks

    def _create_model(self) -> DLASeg:
        self.logger.info(
            "FairMOT model loaded with the following config:\n\t"
            f"Model type: {self.model_type}\n\t"
            f"Input resolution: {self.input_size}"
            f"Score threshold: {self.score_threshold}\n\t"
            f"Max number of output objects: {self.max_per_image}\n\t"
            f"Min bounding box area: {self.min_box_area}\n\t"
            f"Track buffer: {self.track_buffer}\n\t"
        )
        return self._load_model_weights()

    def _load_model_weights(self) -> DLASeg:
        if not self.model_path.is_file():
            raise ValueError(
                f"Model file does not exist. Please check that {self.model_path} exists."
            )

        ckpt = torch.load(str(self.model_path), map_location="cpu")
        model = DLASeg(self.heads, self.down_ratio)
        model.load_state_dict(ckpt["state_dict"], strict=False)
        model.to(self.device).eval()
        return model

    def _preprocess(self, image: np.ndarray) -> np.ndarray:
        """Preprocesses the input image by padded resizing with letterbox and
        normalising RGB values.

        Args:
            image (np.ndarray): Input video frame.

        Returns:
            (np.ndarray): Preprocessed image.
        """
        # Padded resize
        padded_image = letterbox(
            image, height=self.input_size[1], width=self.input_size[0]
        )
        # Normalize RGB
        padded_image = padded_image[..., ::-1].transpose(2, 0, 1)
        padded_image = np.ascontiguousarray(padded_image, dtype=np.float32)
        padded_image /= 255.0

        return padded_image

    @staticmethod
    def _postprocess(tlwhs: np.ndarray, image_shape: Tuple[int, ...]) -> np.ndarray:
        """Post-processes detection bounding boxes by converting them from
        [t, l, w, h] to normalized [x1, y1, x2, y2] format which is required by
        other PeekingDuck draw nodes. (t, l) is the top-left corner, w is
        width, and h is height. (x1, y1) is the top-left corner and (x2, y2) is
        the bottom-right corner.

        Args:
            tlwhs (np.ndarray): Bounding boxes in [t, l, w, h] format.
            image_shape (Tuple[int, ...]): Dimensions of the original video
                frame.

        Returns:
            (np.ndarray): Bounding boxes in normalized [x1, y1, x2, y2] format.
        """
        return tlwh2xyxyn(tlwhs, *image_shape)


def _combine_stracks(stracks_1: List[STrack], stracks_2: List[STrack]) -> List[STrack]:
    """Combines two list of STrack together.

    Args:
        stracks_1 (List[STrack]): List of STrack.
        stracks_2 (List[STrack]): List of STrack.

    Returns:
        (List[STrack]): Combined list of STrack.
    """
    stracks = {track.track_id: track for track in stracks_1}
    for track in stracks_2:
        tid = track.track_id
        if tid not in stracks:
            stracks[tid] = track
    return list(stracks.values())


def _remove_duplicate_stracks(
    stracks_1: List[STrack], stracks_2: List[STrack]
) -> Tuple[List[STrack], List[STrack]]:
    """Remove duplicate STrack based on costs computed using
    Intersection-over-Union (IoU) values. Duplicates are identified by
    cost<0.15, the STrack that is more recently created is marked as
    the duplicate.

    Args:
        stracks_1 (List[STrack]): List of STrack.
        stracks_2 (List[STrack]): List of STrack.

    Returns:
        (Tuple[List[STrack], List[STrack]]): Lists of STrack with duplicates
            removed.
    """
    distances = matching.iou_distance(stracks_1, stracks_2)
    pairs = np.where(distances < 0.15)
    duplicates_1 = []
    duplicates_2 = []
    for idx_1, idx_2 in zip(*pairs):
        age_1 = stracks_1[idx_1].frame_id - stracks_1[idx_1].start_frame
        age_2 = stracks_2[idx_2].frame_id - stracks_2[idx_2].start_frame
        if age_1 > age_2:
            duplicates_2.append(idx_2)
        else:
            duplicates_1.append(idx_1)
    return (
        [t for i, t in enumerate(stracks_1) if i not in duplicates_1],
        [t for i, t in enumerate(stracks_2) if i not in duplicates_2],
    )


def _subtract_stracks(stracks_1: List[STrack], stracks_2: List[STrack]) -> List[STrack]:
    """Removes stracks_2 from stracks_1.

    Args:
        stracks_1 (List[STrack]): List of STrack.
        stracks_2 (List[STrack]): List of STrack.

    Returns:
        (List[STrack]): List of STrack.
    """
    stracks = {track.track_id: track for track in stracks_1}
    for track in stracks_2:
        tid = track.track_id
        if tid in stracks:
            del stracks[tid]
    return list(stracks.values())



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🕺 Fast Pose Estimation model."""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.abstract_node import AbstractNode
from peekingduck.pipeline.nodes.model.posenetv1 import posenet_model


class Node(AbstractNode):
    """Initializes a PoseNet model to detect human poses from an image.

    The PoseNet node is capable of detecting multiple human figures
    simultaneously per inference and for each detected human figure, 17
    keypoints are estimated. The keypoint indices table can be found
    :ref:`here <whole-body-keypoint-ids>`.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |keypoints_data|

        |keypoint_scores_data|

        |keypoint_conns_data|

        |bbox_labels_data|

    Configs:
        model_type (:obj:`Union[str, int]`):
            **{"resnet", 50, 75, 100}, default="resnet"**. |br|
            Defines the backbone model for PoseNet.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        resolution (:obj:`Dict`):
            **default = { height: 225, width: 225 }**. |br|
            Resolution of input array to PoseNet model.
        max_pose_detection (:obj:`int`): **default = 10**. |br|
            Maximum number of poses to be detected.
        score_threshold (:obj:`float`): **[0, 1], default = 0.4**. |br|
            Detected keypoints confidence score threshold, only keypoints above
            threshold will be kept in output.

    References:
        PersonLab: Person Pose Estimation and Instance Segmentation with a
        Bottom-Up, Part-Based, Geometric Embedding Model:
        https://arxiv.org/abs/1803.08225

        Code adapted from https://github.com/rwightman/posenet-python
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = posenet_model.PoseNetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads the image input and returns the bboxes of the specified
        objects chosen to be detected.
        """
        bboxes, keypoints, keypoint_scores, keypoint_conns = self.model.predict(
            inputs["img"]
        )
        bbox_labels = np.array(["person"] * len(bboxes))
        bboxes = np.clip(bboxes, 0, 1)

        outputs = {
            "bboxes": bboxes,
            "bbox_labels": bbox_labels,
            "keypoints": keypoints,
            "keypoint_scores": keypoint_scores,
            "keypoint_conns": keypoint_conns,
        }
        return outputs



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Deformable ConvNets v2 in PyTorch."""

import math
from typing import Tuple, Union

import torch
import torchvision.ops
from torch import nn
from torch.nn.modules.utils import _pair


class DCNv2(nn.Module):
    """Deformable ConvNets v2 as described in
    `Deformable ConvNets v2: More Deformable, Better Results
    <https://arxiv.org/abs/1811.11168>`__.

    Attributes:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
        kernel_size (Tuple[int, int]): Size of the convolving kernel.
        stride (Tuple[int, int]): Stride of the convolution.
        padding (Tuple[int, int]): Padding added to all four sides of the
            input.
        dilation (Tuple[int, int]): Spacing between kernel elements.
        deformable_groups (int): Used to determine the number of offset groups.
        weight (torch.nn.Parameter): Convolution weights.
        bias (torch.nn.Parameter): Bias terms for convolution.
        conv_offset_mask (torch.nn.Conv2d): 2D convolution to generate the
            offset and mask.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
        kernel_size (Union[int, Tuple[int, int])): Size of the convolving
            kernel.
        stride (Union[int, Tuple[int, int]]): Stride of the convolution.
        padding (Union[int, Tuple[int, int]]): Padding added to all four sides
            of the input.
        dilation (Union[int, Tuple[int, int]]): Spacing between kernel
            elements.
        deformable_groups (int): Used to determine the number of offset groups.

    Reference:
        PyTorch-Deformable-Convolution-v2:
        https://github.com/developer0hye/PyTorch-Deformable-Convolution-v2

        Deformable Convolutional Networks V2 with Pytorch 1.0:
        https://github.com/ifzhang/DCNv2/tree/pytorch_1.7
    """

    # pylint: disable=redefined-builtin, too-many-instance-attributes

    num_chunks = 3  # Num channels for offset + mask

    def __init__(  # pylint: disable=too-many-arguments
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride: Union[int, Tuple[int, int]],
        padding: Union[int, Tuple[int, int]],
        dilation: Union[int, Tuple[int, int]] = 1,
        deformable_groups: int = 1,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.stride = _pair(stride)
        self.padding = _pair(padding)
        self.dilation = _pair(dilation)
        self.deformable_groups = deformable_groups

        self.weight = nn.Parameter(
            torch.Tensor(out_channels, in_channels, *self.kernel_size)
        )
        self.bias = nn.Parameter(torch.Tensor(out_channels))
        self.reset_parameters()

        num_offset_mask_channels = (
            self.deformable_groups
            * self.num_chunks
            * self.kernel_size[0]
            * self.kernel_size[1]
        )
        self.conv_offset_mask = nn.Conv2d(
            self.in_channels,
            num_offset_mask_channels,
            self.kernel_size,
            self.stride,
            self.padding,
            bias=True,
        )
        self.init_offset()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            input (torch.Tensor): Input from the previous layer.

        Returns:
            (torch.Tensor): Result of convolution.
        """
        out = self.conv_offset_mask(input)
        offset_1, offset_2, mask = torch.chunk(out, self.num_chunks, dim=1)
        offset = torch.cat((offset_1, offset_2), dim=1)
        mask = torch.sigmoid(mask)

        return torchvision.ops.deform_conv2d(
            input=input,
            offset=offset,
            weight=self.weight,
            bias=self.bias,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            mask=mask,
        )

    def init_offset(self) -> None:
        """Initializes the weight and bias for `conv_offset_mask`."""
        self.conv_offset_mask.weight.data.zero_()
        if self.conv_offset_mask.bias is not None:
            self.conv_offset_mask.bias.data.zero_()

    def reset_parameters(self) -> None:
        """Re-initialize parameters using a method similar to He
        initialization with mode='fan_in' and gain=1.
        """
        fan_in = self.in_channels
        for k in self.kernel_size:
            fan_in *= k
        std = 1.0 / math.sqrt(fan_in)
        self.weight.data.uniform_(-std, std)
        self.bias.data.zero_()



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions used by tracking-by-detection trackers."""

import numpy as np


def iou_candidates(bbox: np.ndarray, candidates: np.ndarray) -> np.ndarray:
    """Computes Intersection-over-Union of `bbox` and each of the candidate
    bbox in `candidates`.

    Args:
        bbox (np.ndarray): A bounding box in format `(top left x, top left y,
            width, height)`.
        candidates (np.ndarray): A matrix of candidate bounding boxes
            (one per row) in the same format as `bbox`.
    Returns:
        np.ndarray: The IoU in [0, 1] between the `bbox` and each candidate. A
            higher score means a larger fraction of the `bbox` is occluded by
            the candidate.
    """
    bbox_tl = bbox[:2]
    bbox_br = bbox[:2] + bbox[2:]
    candidates_tl = candidates[:, :2]
    candidates_br = candidates[:, :2] + candidates[:, 2:]

    top_left = np.c_[
        np.maximum(bbox_tl[0], candidates_tl[:, 0])[:, np.newaxis],
        np.maximum(bbox_tl[1], candidates_tl[:, 1])[:, np.newaxis],
    ]
    bottom_right = np.c_[
        np.minimum(bbox_br[0], candidates_br[:, 0])[:, np.newaxis],
        np.minimum(bbox_br[1], candidates_br[:, 1])[:, np.newaxis],
    ]
    width_height = np.maximum(0.0, bottom_right - top_left)

    area_intersection = width_height.prod(axis=1)
    area_bbox = bbox[2:].prod()
    area_candidates = candidates[:, 2:].prod(axis=1)

    return area_intersection / (area_bbox + area_candidates - area_intersection)


def iou_tlwh(bbox_1: np.ndarray, bbox_2: np.ndarray) -> float:
    """Calculates the Intersection-over-Union (IoU) of two bounding boxes. Each
    bounding box as the format (t, l, w, h) where (t, l) is the top-left
    corner, w is the width, and h is the height.

    Args:
        bbox_1 (np.ndarray): The first bounding box.
        bbox_2 (np.ndarray): The other bounding box.

    Returns:
        (float): IoU of bbox1, bbox2.
    """
    bbox_1 = bbox_1.astype(float)
    bbox_2 = bbox_2.astype(float)
    overlap_x1 = np.maximum(bbox_1[0], bbox_2[0])
    overlap_y1 = np.maximum(bbox_1[1], bbox_2[1])
    overlap_x2 = np.minimum(bbox_1[0] + bbox_1[2], bbox_2[0] + bbox_2[2])
    overlap_y2 = np.minimum(bbox_1[1] + bbox_1[3], bbox_2[1] + bbox_2[3])

    overlap_w = max(0.0, overlap_x2 - overlap_x1)
    overlap_h = max(0.0, overlap_y2 - overlap_y1)
    if overlap_w <= 0 or overlap_h <= 0:
        return 0.0

    area_intersection = overlap_w * overlap_h
    area_bbox_1 = bbox_1[2:].prod()
    area_bbox_2 = bbox_2[2:].prod()

    return area_intersection / (area_bbox_1 + area_bbox_2 - area_intersection)


def xyxyn2tlwh(inputs: np.ndarray, height: int, width: int) -> np.ndarray:
    """Converts bounding boxes format from (x1, y2, x2, y2) to (t, l, w, h).
    (x1, y1) is the normalized coordinates of the top-left corner, (x2, y2) is
    the normalized coordinates of the bottom-right corner. (t, l) is the
    original coordinates of the top-left corner, (w, h) is the original width
    and height of the bounding box.

    Args:
        inputs (np.ndarray): Bounding box coordinates with (x1, y1, x2, y2)
            format.
        height (int): Original height of bounding box.
        width (int): Original width of bounding box.

    Returns:
        (np.ndarray): Converted bounding box coordinates with (t, l, w, h)
            format.
    """
    outputs = np.empty_like(inputs)
    outputs[:, 0] = inputs[:, 0] * width  # Bottom left x
    outputs[:, 1] = inputs[:, 1] * height  # Bottom left y
    outputs[:, 2] = (inputs[:, 2] - inputs[:, 0]) * width  # Top right x
    outputs[:, 3] = (inputs[:, 3] - inputs[:, 1]) * height  # Top right y
    return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Object detection class using yolo model to find object bboxes
"""


import logging
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import numpy as np
import tensorflow as tf

from peekingduck.utils.graph_functions import load_graph


class Detector:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """Object detection class using YOLO model to find object bboxes"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        model_dir: Path,
        class_names: List[str],
        detect_ids: List[int],
        model_type: str,
        model_file: Dict[str, str],
        model_nodes: Dict[str, List[str]],
        max_output_size_per_class: int,
        max_total_size: int,
        input_size: int,
        iou_threshold: float,
        score_threshold: float,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.class_names = class_names
        self.model_type = model_type
        self.model_path = model_dir / model_file[self.model_type]
        self.model_nodes = model_nodes

        self.max_output_size_per_class = max_output_size_per_class
        self.max_total_size = max_total_size
        self.input_size = (input_size, input_size)
        self.iou_threshold = iou_threshold
        self.score_threshold = score_threshold

        self.detect_ids = detect_ids
        self.yolo = self._create_yolo_model()

    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Detect all objects' bounding box from one image

        Args:
            image (np.ndarray): input image

        Return:
            boxes (np.array): an array of bounding box with definition like
                (x1, y1, x2, y2), in a coordinate system with original point in
                the top-left corner
        """
        image = self._preprocess(image)

        pred = self.yolo(image)[-1]

        bboxes, scores, classes = self._postprocess(pred[:, :, :4], pred[:, :, 4:])
        labels = np.array([self.class_names[int(i)] for i in classes])

        return bboxes, labels, scores

    def _create_yolo_model(self) -> Callable:
        """Creates YOLO model for human detection."""
        self.logger.info(
            "YOLO model loaded with following configs: \n\t"
            f"Model type: {self.model_type}, \n\t"
            f"Input resolution: {self.input_size}, \n\t"
            f"IDs being detected: {self.detect_ids} \n\t"
            f"Max detections per class: {self.max_output_size_per_class}, \n\t"
            f"Max total detections: {self.max_total_size}, \n\t"
            f"IOU threshold: {self.iou_threshold}, \n\t"
            f"Score threshold: {self.score_threshold}"
        )

        return self._load_yolo_weights()

    def _load_yolo_weights(self) -> Callable:
        """When loading a graph model, you need to explicitly state the input
        and output nodes of the graph. It is usually x:0 for input and Identity:0
        for outputs, depending on how many output nodes you have.
        """
        if not self.model_path.is_file():
            raise ValueError(
                f"Graph file does not exist. Please check that {self.model_path} exists"
            )
        return load_graph(
            str(self.model_path),
            inputs=self.model_nodes["inputs"],
            outputs=self.model_nodes["outputs"],
        )

    def _postprocess(
        self, pred_boxes: tf.Tensor, pred_scores: tf.Tensor
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        bboxes, scores, classes, valid_dets = tf.image.combined_non_max_suppression(
            tf.reshape(pred_boxes, (tf.shape(pred_boxes)[0], -1, 1, 4)),
            tf.reshape(
                pred_scores, (tf.shape(pred_scores)[0], -1, tf.shape(pred_scores)[-1])
            ),
            self.max_output_size_per_class,
            self.max_total_size,
            self.iou_threshold,
            self.score_threshold,
        )
        num_valid = valid_dets[0]

        classes = classes.numpy()[0]
        classes = classes[:num_valid]
        # only identify objects we are interested in
        mask = np.isin(classes, self.detect_ids)
        classes = classes[mask]

        scores = scores.numpy()[0]
        scores = scores[:num_valid]
        scores = scores[mask]

        bboxes = bboxes.numpy()[0]
        bboxes = bboxes[:num_valid]
        bboxes = bboxes[mask]

        # swapping x and y axes
        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]

        return bboxes, scores, classes

    def _preprocess(self, image: np.ndarray) -> tf.Tensor:
        processed_image = tf.convert_to_tensor(image.astype(np.float32))
        processed_image = tf.expand_dims(processed_image, 0)
        processed_image = tf.image.resize(processed_image, self.input_size) / 255.0

        return processed_image
